"""This module contains the ahbconditions class."""

import json
import re
from pathlib import Path

from docx.table import Table as DocxTable
from efoli import EdifactFormat
from pydantic import BaseModel, ConfigDict, Field

from kohlrahbi.logger import logger


class AhbConditions(BaseModel):
    """
    Class which contains a dict of conditions for each edifact format
    """

    conditions_dict: dict[EdifactFormat, dict[str, str]] = Field(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_docx_table(cls, docx_tables: list[DocxTable], edifact_format: EdifactFormat) -> "AhbConditions":
        """
        Create an AhbPackageTable object from a docx table.
        """
        table_data = []
        for table in docx_tables:
            for row in table.rows:
                if row.cells[-1].text and row.cells[0].text != "EDIFACT Struktur":
                    row_data = row.cells[-1].text
                    table_data.append(row_data)

        conditions_dict = {}
        are_there_conditions = len(table_data) > 0
        if are_there_conditions:
            conditions_dict = AhbConditions.collect_conditions(
                conditions_list=table_data, edifact_format=edifact_format
            )

        return cls(conditions_dict=conditions_dict)

    @staticmethod
    def collect_conditions(
        conditions_list: list[str], edifact_format: EdifactFormat
    ) -> dict[EdifactFormat, dict[str, str]]:
        """collect conditions from list of all conditions and store them in conditions dict."""
        conditions_dict: dict[EdifactFormat, dict[str, str]] = {edifact_format: {}}

        conditions_str = "".join(conditions_list)
        conditions_dict = parse_conditions_from_string(conditions_str, edifact_format, conditions_dict)
        logger.info("The package conditions for %s were collected.", edifact_format)
        return conditions_dict

    def include_condition_dict(self, to_add: dict[EdifactFormat, dict[str, str]] | None) -> None:
        """ " Include a dict of conditions to the conditions_dict"""
        if to_add is None:
            logger.info("Conditions dict to be added is empty.")
            return
        for edifact_format, edi_cond_dict in to_add.items():
            for condition_key, condition_text in edi_cond_dict.items():
                if edifact_format in self.conditions_dict:  # pylint:disable=unsupported-membership-test
                    if (
                        # pylint:disable=unsubscriptable-object
                        condition_key in self.conditions_dict[edifact_format]
                        and len(condition_text) > len(self.conditions_dict[edifact_format][condition_key])
                        # pylint:disable=unsubscriptable-object
                        or condition_key not in self.conditions_dict[edifact_format]
                    ):
                        self.conditions_dict[edifact_format][condition_key] = condition_text
                else:
                    self.conditions_dict[edifact_format] = {  # pylint:disable=unsupported-assignment-operation
                        condition_key: condition_text
                    }

        logger.info("Conditions were updated.")

    def dump_as_json(self, output_directory_path: Path) -> None:
        """
        Writes all collected conditions to a json file.
        The file will be stored in the directory:
            'output_directory_path/<edifact_format>/conditions.json'
        """
        # pylint:disable=no-member
        for edifact_format, format_cond_dict in self.conditions_dict.items():
            condition_json_output_directory_path = output_directory_path / str(edifact_format)
            condition_json_output_directory_path.mkdir(parents=True, exist_ok=True)
            file_path = condition_json_output_directory_path / "conditions.json"
            # resort ConditionKeyConditionTextMappings for output
            sorted_condition_dict = {k: format_cond_dict[k] for k in sorted(format_cond_dict, key=int)}
            array = [
                {"condition_key": i, "condition_text": sorted_condition_dict[i], "edifact_format": edifact_format}
                for i in sorted_condition_dict
            ]
            with open(file_path, "w", encoding="utf-8") as file:
                json.dump(array, file, ensure_ascii=False, indent=2)

            logger.info(
                "The conditions.json file for %s is saved at %s",
                edifact_format,
                file_path,
            )


def parse_conditions_from_string(
    conditions_text: str, edifact_format: EdifactFormat, conditions_dict: dict[EdifactFormat, dict[str, str]]
) -> dict[EdifactFormat, dict[str, str]]:
    """
    Takes string with some conditions and sorts it into a dict.
    """
    # Split the input into parts enclosed in square brackets and other parts
    matches = re.findall(
        r"\[(\d+)](.*?)(?=\[\d+]|$)",
        conditions_text,
        re.DOTALL,
    )
    for match in matches:
        # make text prettier:
        text = match[1].strip()
        text = re.sub(r"\s+", " ", text)

        # check whether condition was already collected:
        existing_text = conditions_dict[edifact_format].get(match[0])
        is_condition_key_collected_yet = existing_text is not None
        key_exits_but_shorter_text = existing_text is not None and len(text) > len(existing_text)
        if not is_condition_key_collected_yet or key_exits_but_shorter_text:
            conditions_dict[edifact_format][match[0]] = text
    return conditions_dict



"""
This module contains the DocxFileFinder class.
"""

import re
from itertools import groupby
from pathlib import Path

from efoli import EdifactFormat, get_format_of_pruefidentifikator
from pydantic import BaseModel

from kohlrahbi.logger import logger


class EdiEnergyDocument(BaseModel):
    """
    This class represents an EDI Energy document.
    """

    filename: Path
    document_version: str
    version_major: int
    version_minor: int
    version_suffix: str
    valid_from: int
    valid_until: int

    @classmethod
    def from_path(cls, path: Path) -> "EdiEnergyDocument":
        """
        Create an EdiEnergyDocument object from a file path.
        """

        file_name = extract_document_version_and_valid_dates(path.name)
        assert file_name is not None, f"Could not extract document version and valid dates from {path.name}."
        return cls(
            filename=path,
            document_version=file_name["document_version"],
            version_major=int(file_name["version_major"]),
            version_minor=int(file_name["version_minor"]),
            version_suffix=file_name["version_suffix"],
            valid_from=int(file_name["valid_from"]),
            valid_until=int(file_name["valid_until"]),
        )

    def __lt__(self, other: "EdiEnergyDocument") -> bool:
        """
        Compare two EdiEnergyDocument instances based on
        their document_version(major, minor and suffix), valid_until, and valid_from.

        I did not know how the tuple comparison works in Python, so I looked it up:

        Python compares tuples lexicographically, meaning it compares the elements one by one from left to right.
        The comparison starts with the first elements of both tuples:
          If self.valid_from is less than other.valid_from, the entire expression evaluates to True.
          If self.valid_from is greater than other.valid_from, the entire expression evaluates to False.
          If self.valid_from is equal to other.valid_from, Python moves to the next elements in the tuples.
        This process continues with self.valid_until vs. other.valid_until and then with the version numbers.

        Args:
            other (EdiEnergyDocument): The other document to compare against.

        Returns:
            bool: True if this document is considered less than the other document, False otherwise.
        """
        return (self.valid_from, self.valid_until, self.version_major, self.version_minor, self.version_suffix) < (
            other.valid_from,
            other.valid_until,
            other.version_major,
            other.version_minor,
            other.version_suffix,
        )


def extract_document_version_and_valid_dates(
    filename: str,
) -> dict[str, str] | None:
    """Extract the document version and valid dates from the filename.

    Parameters:
    - filename (str): The filename of the document.

    Returns:
    - tuple[str, str, str]: A tuple containing the document version, valid from date, and valid until date.
    """

    # Pattern to extract detailed version number, valid until and valid from dates
    document_name_pattern = re.compile(
        r"-informatorischeLesefassung"
        r"(?P<document_version>(?:S|G)?(?P<version_major>\d+)\.(?P<version_minor>\d+)(?P<version_suffix>[a-z]?))"
        r"(?:_|KonsolidierteLesefassung|-AußerordentlicheVeröffentlichung)?"
        r"([A-Za-z0-9.]+)?"
        r"_(?P<valid_until>\d{8})_(?P<valid_from>\d{8})\.docx$",
        re.IGNORECASE,
    )
    matches = document_name_pattern.search(filename)
    try:
        if matches:
            return matches.groupdict()
    except ValueError as e:
        logger.error("Error extracting document version and valid dates: %s", e)
    return None


def get_most_recent_file(group_items: list[Path]) -> Path | None:
    """
    Find the most recent file in a group of files based on specific criteria.

    Parameters:
    - group_items (List[Path]): A list of Path objects representing the file paths.

    Returns:
    - Path: A Path object representing the most recent file.
    """

    try:
        # Define the keywords to filter relevant files
        keywords = ["konsolidiertelesefassungmitfehlerkorrekturen", "außerordentlicheveröffentlichung"]
        files_containing_keywords = [
            path for path in group_items if any(keyword in path.name.lower() for keyword in keywords)
        ]
        if any(files_containing_keywords):
            list_of_edi_energy_documents = [EdiEnergyDocument.from_path(path) for path in files_containing_keywords]
        else:
            list_of_edi_energy_documents = [EdiEnergyDocument.from_path(path) for path in group_items]
        most_recent_file = max(list_of_edi_energy_documents)

        return most_recent_file.filename

    except ValueError as e:

        logger.error("Error processing group items: %s", e)
    return None


class DocxFileFinder(BaseModel):
    """
    This class is responsible for finding the docx files in the input directory.
    It can find MIG and AHB docx files.
    """

    paths_to_docx_files: list[Path]

    @classmethod
    def from_input_path(cls, input_path: Path) -> "DocxFileFinder":
        """
        Create an DocxFileFinder object from the input path.
        """

        ahb_file_paths: list[Path] = [path for path in input_path.iterdir() if path.is_file() if path.suffix == ".docx"]
        if not any(ahb_file_paths):  # this is suspicious at least
            logger.warning("The directory '%s' does not contain any docx files.", input_path.absolute())
        return cls(paths_to_docx_files=ahb_file_paths)

    @staticmethod
    def get_first_part_of_ahb_docx_file_name(path_to_ahb_document: Path) -> str:
        """
        Return the first part of the AHB docx file name.
        The first part contains the information about the EDIFACT formats.
        """

        return path_to_ahb_document.name.split("-")[0]

    def filter_for_latest_ahb_docx_files(self) -> None:
        """
        Filter the list of AHB docx paths for the latest AHB docx files.
        The latest files contain `LesefassungmitFehlerkorrekturen` in their file names.
        This method is _not_ pure. It changes the state of the object.
        """
        self.paths_to_docx_files = self.filter_ahb_docx_files(self.paths_to_docx_files)
        grouped_files = self.group_files_by_name_prefix(self.paths_to_docx_files)
        self.paths_to_docx_files = self.filter_latest_version(grouped_files)

    @staticmethod
    def filter_ahb_docx_files(paths_to_docx_files: list[Path]) -> list[Path]:
        """
        This function filters the docx files which contain the string "AHB" in their file name.
        """
        return [path for path in paths_to_docx_files if "AHB" in path.name]

    @staticmethod
    def filter_for_docx_files_with_change_history(paths_to_docx_files: list[Path]) -> list[Path]:
        """
        This function filters the docx files which contain a change history.
        At this time it seems that all docx files have a change history.
        But this may change in the future, so search for some keywords in the file name.
        """
        return [
            path
            for path in paths_to_docx_files
            if "AHB" in path.name
            or "MIG" in path.name
            or "AllgemeineFestlegungen" in path.name
            or "Codeliste" in path.name
            or "Entscheidungsbaum" in path.name
        ]

    # pylint: disable=line-too-long
    @staticmethod
    def group_files_by_name_prefix(paths_to_docx_files: list[Path]) -> dict[str, list[Path]]:
        """
        This function groups the docx files by their name prefix.
        Groups may now look like this:
        {'APERAKCONTRLAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/APERAKCONTRLAHB-informatorischeLesefassung2.3m_99991231_20231001.docx')],
         'COMDISAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/COMDISAHB-informatorischeLesefassung1.0dKonsolidierteLesefassungmitFehlerkorrekturenStand20.07.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/COMDISAHB-informatorischeLesefassung1.0d_99991231_20231001.docx')],
         'HerkunftsnachweisregisterAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/HerkunftsnachweisregisterAHB-informatorischeLesefassung2.3cKonsolidierteLesefassungmitFehlerkorrekturenStand19.06.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/HerkunftsnachweisregisterAHB-informatorischeLesefassung2.3c_99991231_20231001.docx')],
         'IFTSTAAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/IFTSTAAHB-informatorischeLesefassung2.0eKonsolidierteLesefassungmitFehlerkorrekturenStand20.07.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/IFTSTAAHB-informatorischeLesefassung2.0e_99991231_20231001.docx')],
         'INVOICREMADVAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/INVOICREMADVAHB-informatorischeLesefassung2.5bKonsolidierteLesefassungmitFehlerkorrekturenStand20.07.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/INVOICREMADVAHB-informatorischeLesefassung2.5b_99991231_20231001.docx')],
         'MSCONSAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/MSCONSAHB-informatorischeLesefassung3.1c_99991231_20231001.docx')],
         'ORDERSORDRSPAHBMaBiS': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/ORDERSORDRSPAHBMaBiS-informatorischeLesefassung2.2c_99991231_20231001.docx')],
         'PARTINAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/PARTINAHB-informatorischeLesefassung1.0c_99991231_20231001.docx')],
         'PRICATAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/PRICATAHB-informatorischeLesefassung2.0c_99991231_20231001.docx')],
         'REQOTEQUOTESORDERSORDRSPORDCHGAHB': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/REQOTEQUOTESORDERSORDRSPORDCHGAHB-informatorischeLesefassung2.2KonsolidierteLesefassungmitFehlerkorrekturenStand20.07.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/REQOTEQUOTESORDERSORDRSPORDCHGAHB-informatorischeLesefassung2.2_99991231_20231001.docx')],
         'UTILMDAHBGas': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILMDAHBGas-informatorischeLesefassung1.0aKonsolidierteLesefassungmitFehlerkorrekturenStand29.06.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILMDAHBGas-informatorischeLesefassung1.0a_99991231_20231001.docx')],
         'UTILMDAHBMaBiS': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILMDAHBMaBiS-informatorischeLesefassung4.1_99991231_20231001.docx')],
         'UTILMDAHBStrom': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILMDAHBStrom-informatorischeLesefassung1.1KonsolidierteLesefassungmitFehlerkorrekturenStand29.06.2023_99991231_20231001.docx'), WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILMDAHBStrom-informatorischeLesefassung1.1_99991231_20231001.docx')],
         'UTILTSAHBBerechnungsformel': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILTSAHBBerechnungsformel-informatorischeLesefassung1.0e_99991231_20231001.docx')], 'UTILTSAHBDefinitionen': [WindowsPath('../edi_energy_mirror/edi_energy_de/future/UTILTSAHBDefinitionen-informatorischeLesefassung1.1_99991231_20231001.docx')]}
        """
        return {
            group_key: list(group)
            for group_key, group in groupby(
                sorted(paths_to_docx_files, key=DocxFileFinder.get_first_part_of_ahb_docx_file_name),
                key=DocxFileFinder.get_first_part_of_ahb_docx_file_name,
            )
        }

    @staticmethod
    def filter_latest_version(groups: dict[str, list[Path]]) -> list[Path]:
        """
        Filters and returns the latest version of the AHB or MIG .docx files
        from the provided groups based on specific criteria.

        The latest version is determined based on the presence of specific
        keywords in the filename and the numerical suffix in the filename.

        Parameters:
        - groups (Dict[str, List[Path]]): A dictionary where keys are group identifiers
        and values are lists of Path objects representing the file paths.

        Returns:
        - List[Path]: A list of Path objects representing the latest version of the files.
        """
        result: list[Path] = []

        for group_items in groups.values():
            most_recent_file = get_most_recent_file(group_items)
            assert most_recent_file is not None, "Could not find the most recent file."
            result.append(most_recent_file)
        return result

    def filter_for_latest_mig_and_ahb_docx_files(self) -> None:
        """
        Filter the list of MIG docx paths for the latest MIG docx files.
        """
        self.paths_to_docx_files = self.filter_for_docx_files_with_change_history(self.paths_to_docx_files)
        grouped_files = self.group_files_by_name_prefix(self.paths_to_docx_files)
        self.paths_to_docx_files = self.filter_latest_version(grouped_files)

    def filter_docx_files_for_edifact_format(self, edifact_format: EdifactFormat) -> None:
        """
        Returns a list of docx files which contain the given edifact format.
        This method is not pure. It changes the state of the object.
        """

        self.paths_to_docx_files = [path for path in self.paths_to_docx_files if str(edifact_format) in path.name]

    def remove_temporary_files(self) -> None:
        """
        This method removes all temporary files from paths_to_docx_files.
        Temporary files lead to the exception `BadZipFile: File is not a zip file`.
        It appears if a docx file is opened by Word.
        """

        self.paths_to_docx_files = [path for path in self.paths_to_docx_files if not path.name.startswith("~")]

    def get_docx_files_which_may_contain_searched_pruefi(self, searched_pruefi: str) -> list[Path]:
        """
        This functions takes a pruefidentifikator and returns a list of docx files which can contain the searched pruefi
        Unfortunately, it is not clear in which docx the pruefidentifikator you are looking for is located.
        A 11042 belongs to the UTILMD format. However, there are seven docx files that describe the UTILMD format.
        A further reduction of the number of files is not possible with the pruefidentifikator only.
        This method is _not_ pure. It changes the state of the object.
        """

        edifact_format = get_format_of_pruefidentifikator(searched_pruefi)
        if edifact_format is None:
            logger.exception("❌ There is no known format for the prüfi '%s'.", searched_pruefi)
            raise ValueError(f"There is no known format for the prüfi '{searched_pruefi}'.")

        self.filter_for_latest_ahb_docx_files()
        self.filter_docx_files_for_edifact_format(edifact_format=edifact_format)
        if (
            edifact_format == EdifactFormat.UTILMD
            and searched_pruefi.startswith("11")
            and all("202310" in path.name for path in self.paths_to_docx_files)
        ):
            logger.info(
                # pylint:disable=line-too-long
                "You searched for a UTILMD prüfi %s starting with the soon deprecated prefix '11' but all relevant files %s are valid from 2023-10 onwards. They won't contain any match.",
                searched_pruefi,
                ", ".join([path.name for path in self.paths_to_docx_files]),
            )
            return []
        return self.paths_to_docx_files

    def get_all_docx_files_which_contain_change_histories(self) -> list[Path]:
        """
        This function returns a list of docx files which probably contain a change history.
        Only format documents like UTILMD, MSCONS etc. contain a change history.
        """

        self.filter_for_latest_mig_and_ahb_docx_files()
        self.remove_temporary_files()

        return self.paths_to_docx_files

    def get_docx_files_which_contain_quality_map(self) -> list[Path]:
        """
        This function returns a list of docx files which contain a quality map.
        """

        self.filter_for_latest_ahb_docx_files()
        self.remove_temporary_files()

        indicator_string = "UTILMDAHBStrom"
        self.paths_to_docx_files = [path for path in self.paths_to_docx_files if indicator_string in path.name]

        return self.paths_to_docx_files



"""
This module contains the class AhbTableRow
"""

from typing import Optional

import pandas as pd
from docx.table import _Cell
from pydantic import BaseModel, ConfigDict

from kohlrahbi.docxtablecells import BedingungCell, BodyCell, EdifactStrukturCell
from kohlrahbi.enums import RowType
from kohlrahbi.seed import Seed


# pylint:disable=too-few-public-methods
class AhbTableRow(BaseModel):
    """
    A AhbTableRow is a single row from an AHB table.
    It contains a seed and the three cells (columns).
    """

    seed: Seed
    edifact_struktur_cell: _Cell
    middle_cell: _Cell
    bedingung_cell: _Cell

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def parse(self, row_type: RowType) -> Optional[pd.DataFrame]:
        """
        Writes the current row of the current table into the DataFrame depending on the type of the row.
        If the row is a header row, it will be skipped and None will be returned.
        """

        if row_type is RowType.HEADER:
            # we skip the header rows because we scraped it already and there are no new information
            return None

        ahb_row_dataframe = pd.DataFrame(
            # pylint: disable=no-member
            columns=self.seed.column_headers,
            dtype="str",
        )
        # pylint: disable=unsubscriptable-object, no-member
        empty_row: pd.Series = pd.Series(  # type:ignore[type-arg]
            len(ahb_row_dataframe.columns) * [""], index=self.seed.column_headers
        )

        ahb_row_dataframe = pd.concat([ahb_row_dataframe, empty_row.to_frame().T], ignore_index=True)

        # EDIFACT STRUKTUR
        esc: EdifactStrukturCell = EdifactStrukturCell(
            # pylint: disable=no-member
            table_cell=self.edifact_struktur_cell,
            edifact_struktur_cell_left_indent_position=self.seed.edifact_struktur_left_indent_position,
        )
        ahb_row_dataframe = esc.parse(ahb_row_dataframe=ahb_row_dataframe)

        # BODY
        boc: BodyCell = BodyCell(
            # pylint: disable=no-member
            table_cell=self.middle_cell,
            left_indent_position=self.seed.middle_cell_left_indent_position,
            indicator_tabstop_positions=self.seed.tabstop_positions,
        )
        ahb_row_dataframe = boc.parse(ahb_row_dataframe=ahb_row_dataframe)

        # BEDINGUNG
        bec: BedingungCell = BedingungCell(table_cell=self.bedingung_cell)
        ahb_row_dataframe = bec.parse(ahb_row_dataframe)

        return ahb_row_dataframe



# pylint:disable=too-few-public-methods
"""
EDIFACT components are data structures on different hierarchical levels inside an EDIFACT message.
Components contain not only EDIFACT composites but also segments and segment groups.
"""
import re
from abc import ABC
from enum import Enum
from typing import Callable, Iterable, Literal, Mapping, Optional, Union

from pydantic import BaseModel, Field, StringConstraints, field_validator
from typing_extensions import Annotated

from kohlrahbi.models import _check_that_string_is_not_whitespace_or_empty


class DataElementDataType(str, Enum):
    """
    The Data Element Data Type describes with which kind of data element we're dealing with in a data element.
    This information is set but not used anywhere inside MAUS directly but more of a "service" to downstream code.
    """

    TEXT = "TEXT"  #: plain text, e.g. a name
    DATETIME = "DATETIME"  #: a datetime string, usually as RFC3339
    VALUE_POOL = "VALUE_POOL"  #: the user can choose between different possible values


def derive_data_type_from_segment_code(segment_code: str) -> Optional[DataElementDataType]:
    """
    derives the expected data type from the segment code, e.g. `DATETIME` for DTM segments
    :return: The DataType if it can be derived without any doubt, None otherwise
    """
    if segment_code in {"DTM"}:
        return DataElementDataType.DATETIME
    return None


class DataElement(BaseModel, ABC):
    """
    A data element holds specific kinds of data. It is defined in EDIFACT.
    At least for the German energy market communication any data element has a 4 digit key.
    For example in UTILMD the data element that holds the 13 digit market partner ID is data element '3039'
    """

    discriminator: Annotated[Optional[str], StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        None,
        description=(
            "The discriminator uniquely identifies the data element."
            "The discriminator is None if the data element was not found in the MIG."
        ),
    )
    data_element_id: Annotated[str, StringConstraints(strip_whitespace=True, pattern=r"^\d{4}$")] = Field(
        ...,
        description="The ID of the data element (e.g. '0062') for the Nachrichten-Referenznummer",
    )
    value_type: Optional[DataElementDataType] = Field(
        None,
        description=(
            "The value_type allows to describe which type of data we're expecting to be used within this data element."
            "The value_type does not discriminate the type of the data element itself."
        ),
    )


class DataElementFreeText(DataElement):
    """
    A DataElementFreeText is a data element that allows entering arbitrary data.
    This is the main difference to the :class:`DataElementValuePool` which has a finite set of allowed values attached.
    """

    value_type: Literal[DataElementDataType.TEXT] = Field(
        default=DataElementDataType.TEXT,
        description=(
            "The value_type allows to describe which type of data we're expecting to be used within this data element."
            "The value_type does not discriminate the type of the data element itself."
        ),
    )
    ahb_expression: Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        ..., description="Any freetext data element has an ahb expression attached. Could be 'X' but also 'M [13]'"
    )
    free_text: str = Field(
        ...,
        description=(
            "free_text contains the the description of the free text element, e.g. 'Referenz, Identifikation'"
        ),
    )


#: a pattern that matches most of the qualifiers we find in the AHBs
_simple_edifact_qualifier_pattern = re.compile(r"^([A-Z\d]+)|(\d+\.\d+[a-z])$")

#: a pattern that matches the GABi qualifiers: They contain with "-" and lower case "i"/"o"/"n"
gabi_edifact_qualifier_pattern = re.compile(r"^(?!MP-ID)(GABi)?[A-Z\d\-]+(RLM(o|m)T)?$")


def _check_is_edifact_qualifier(value: str) -> str:
    """
    Checks that the given attribute is a valid EDIFACT qualifier.
    Raises a ValueError if not.
    """
    _check_that_string_is_not_whitespace_or_empty(value)
    simple_match = _simple_edifact_qualifier_pattern.match(value)
    if simple_match is not None:
        return value
    gabi_match = gabi_edifact_qualifier_pattern.match(value)
    if gabi_match is not None:
        return value
    raise ValueError(f"The qualifier '{value}' is invalid")


class ValuePoolEntry(BaseModel):
    """
    A value pool entry contains the EDIFACT qualifier, a meaning (German text) and an ahb expression.
    A value pool consists of 1 to n ValuePoolEntries.
    The data element 3055 in UTILMD is a good example for a value pool.
    It is used in the segments NAD+MS and NAD+MR. Its ValuePoolEntries are:
    - (key: "9", meaning: "GS1", ahb_expression: "X")
    - (key: "293", meaning: "DE, BDEW", ahb_expression: "X")
    - (key: "332", meaning: "DE, DVGW", ahb_expression: "X")
    """

    qualifier: Annotated[
        str,
        StringConstraints(strip_whitespace=True, min_length=1),
    ] = Field(..., description="The qualifier in EDIFACT")
    meaning: Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        ...,
        description="The meaning as it is written in the AHB,"
        "e.g. 'Einzug', 'Entwurfs-Version', 'GS1', 'Codeliste Gas G_0057'",
    )
    ahb_expression: Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        ...,
        description="The AHB expression, in most cases this is a simple 'X'; it must not be empty",
    )

    @field_validator("qualifier")
    @classmethod
    def validate_qualifier(cls, v: str) -> str:
        """
        Check that the qualifier is a valid EDIFACT qualifier

        We can not use the pattern parameter of the StringConstraints here, because look-around,
        including look-ahead and look-behind, is not supported
        """
        return _check_is_edifact_qualifier(v)


class DataElementValuePool(DataElement):
    """
    A DataElementValuePool is a data element with a finite set of allowed values.
    These allowed values are referred to as keys of the "value pool".
    The value pool is defined both on MIG level and AHB level
    The set of values allowed according to the AHB is always a subset of the values allowed according to the MIG.
    """

    value_type: Literal[DataElementDataType.VALUE_POOL] = Field(
        default=DataElementDataType.VALUE_POOL, description="Type of the value, if known"
    )
    value_pool: list[ValuePoolEntry] = Field(
        ..., description="The value pool contains at least one value :class:`.ValuePoolEntry`"
    )

    def replace_value_pool(
        self,
        edifact_to_domain_mapping: Mapping[str, str],
        meaning_qualifier_merger: Optional[Callable[[str, str], str]] = None,
    ) -> None:
        """
        If your data model comes from another domain than edifact the value pool from the AHBs, you can, in general, not
        use the same keys. Think e.g. of the Transaktionsgrund in UTILMD. In EDIFACT you might have the values:
        Edifact Domain:
        - E01 = Einzug/Auszug
        - E03 = Wechsel
        - Z33 = Auszug/Stilllegung
        ...
        But if, in your application, you're modelling these values with other values (think e.g. of readable enums):
        Application Domain:
        - EINZUG = Einzug/Auszug
        - WECHSEL = Wechsel
        - STILLLEGUNG = Auszug/Stilllegung
        ...
        You need to transform the value pool from edifact to something that matches the domain from your application.
        To do so, provide this method with an edifact_to_domain_mapping:
        {"E01": "EINZUG","E02": "WECHSEL","E03": "STILLLEGUNG"}
        where each entry represents the mapping of an edifact qualifier to your application domain.

        This method replaces the keys of the ValuePoolEntries if they are found in the mapping.
        By doing so, it allows transforming checks against value pools from the edifact to your application domain.
        This problem does not occur for free text data elements.

        Provide an optional meaning_qualifier_merger that allows you to 'store' the old qualifier in the meaning.
        This is an _unstructured_ way to save the information that is lost with the replacement.
        """
        existing_value_pool_entries: dict[str, ValuePoolEntry] = {x.qualifier: x for x in self.value_pool}
        for existing_value_pool_qualifier, value_pool_entry in existing_value_pool_entries.items():
            if existing_value_pool_qualifier in edifact_to_domain_mapping:
                if meaning_qualifier_merger is not None:
                    value_pool_entry.meaning = meaning_qualifier_merger(
                        value_pool_entry.meaning, existing_value_pool_qualifier
                    )
                value_pool_entry.qualifier = edifact_to_domain_mapping[existing_value_pool_qualifier]

    def has_value_pool_which_is_subset_of(self, entries: Iterable[str]) -> bool:
        """
        returns true iff all qualifiers from the data elements value pool are found in entries
        """
        return all(x.qualifier in entries for x in self.value_pool)


class SegmentLevel(BaseModel, ABC):
    """
    SegmentLevel describes @annika: what does it describe?
    """

    discriminator: str  # no validator here, because it might be None on initialization and will be set later (trust me)
    ahb_expression: Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        ...,
        description="AHB expression",
    )
    ahb_line_index: Optional[int] = Field(
        default=None,
        description=(
            "Allows sorting the segments depending on where they occurred in the FlatAnwendungshandbuch."
            "It won't be serialized though."
        ),
    )

    def reset_ahb_line_index(self) -> None:
        """
        sets the ahb_line_index to None.
        This is to allow comparisons regardless of the index
        """
        self.ahb_line_index = None


class Segment(SegmentLevel):
    """
    A Segment contains multiple data elements.
    """

    data_elements: list[Annotated[Union[DataElementValuePool, DataElementFreeText], Field(discriminator="value_type")]]
    section_name: Optional[str] = Field(
        default=None,
        description=(
            "For the MIG matching it might be necessary to know the section"
            "in which the data element occurred in the AHB."
            "This might be necessary to e.g. distinguish gas and electricity fields which look the same otherwise."
            "See e.g. UTILMD 'Geplante Turnusablesung des MSB (Strom)' vs. 'Geplante Turnusablesung des NB (Gas)'"
        ),
    )
    segment_id: Optional[
        Annotated[str, StringConstraints(strip_whitespace=True, to_upper=True, pattern=r"^\d{5}$")]
    ] = Field(
        default=None,
        description=(
            "The 5 digit segment id,"
            "e.g. '00522' for UTILMD Strom SG12, NAD 'Korrespondenzanschrift des Kunden des Lieferanten'"
        ),
    )

    def get_all_value_pools(self) -> list[DataElementValuePool]:
        """
        find all value pools in this segment
        :return: a list of all value pools
        """
        return [de for de in self.data_elements if isinstance(de, DataElementValuePool)]


class SegmentGroup(SegmentLevel):
    """
    A segment group that contains segments and nested groups.
    On root level of a message there might be a "virtual" segment group of all segments that formally don't have a group
    This group has the key "root".
    """

    segments: Optional[list[Segment]] = Field(default=None, description="The segments inside this very group")
    segment_groups: Optional[list["SegmentGroup"]] = Field(
        default=None, description="Groups that are nested into this group"
    )

    def reset_ahb_line_index(self) -> None:
        self.ahb_line_index = None
        if self.segment_groups:
            for segment_group in self.segment_groups:
                segment_group.reset_ahb_line_index()
        if self.segments:
            for segment in self.segments:
                segment.reset_ahb_line_index()

    def find_segments(self, predicate: Callable[[Segment], bool], search_recursively: bool = True) -> list[Segment]:
        """
        Search for a segment that matches the predicate (in this group and subgroups if 'search_recursively' is set),
        Return results, if found. Return empty list otherwise.
        """
        result: list[Segment] = []
        if self.segments is not None:
            for segment in self.segments:
                if predicate(segment):
                    result.append(segment)
        if search_recursively and self.segment_groups is not None:
            for sub_group in self.segment_groups:
                sub_result = sub_group.find_segments(predicate, search_recursively)
                result += sub_result
        return result

    def get_all_value_pools(self) -> list[DataElementValuePool]:
        """
        recursively find all value pools in this segmentgroup
        :return: a list of all value pools
        """
        result: list[DataElementValuePool] = []
        for segment in self.find_segments(lambda _: True):
            result += segment.get_all_value_pools()
        return result


class EdifactStackLevel(BaseModel):
    """
    The EDIFACT stack level describes the hierarchy level of information inside an EDIFACT message.
    """

    name: str = Field(..., description="The name of the level, e.g. 'Dokument' or 'Nachricht' or 'Meldepunkt'")
    is_groupable: bool = Field(
        ...,
        description=(
            "Describes if this level is groupable / if there are multiple instances of this level"
            "within the same message"
        ),
    )
    index: Optional[int] = Field(default=None, description="The index if present (e.g. 0)")


#: a pattern that matches parts of the json path: https://regex101.com/r/iQzdXK/1
_level_pattern = re.compile(r"\[\"(?P<level_name>[^\[\]]+?)\"\](?:\[(?P<index>\d+)\])?")


class EdifactStack(BaseModel):
    """
    The EdifactStack describes where inside an EDIFACT message data are found.
    The stack is independent of the actual implementation used to create the EDIFACT (be it XML, JSON whatever).
    """

    levels: list[EdifactStackLevel] = Field(..., description="Levels describe the nesting inside an edifact message")

    @staticmethod
    def from_json_path(json_path: str) -> "EdifactStack":
        """
        reads a json path as it is created by "to_json_path" and returns the corresponding edifact stack
        """
        levels: list[EdifactStackLevel] = []
        for level_match in _level_pattern.finditer(json_path):
            level = EdifactStackLevel(name=level_match["level_name"], is_groupable=level_match["index"] is not None)
            if level.is_groupable:
                level.index = int(level_match["index"])
            levels.append(level)
        return EdifactStack(levels=levels)

    def is_sub_stack_of(self, other: "EdifactStack") -> bool:
        """
        Returns true iff this (self) stack is a sub stack of the other provided stack.
        ([Foo][0][Bar]).is_sub_stack_of([Foo][0]) is true.
        """
        if len(other.levels) > len(self.levels):
            # self cannot be a sub path of other if other is "deeper"
            return False
        for level_self, level_other in zip(other.levels, self.levels):  # , strict=False):  # type:ignore[call-overload]
            # strict is False because it's ok if we stop the iteration if self.levels is "exhausted"; explicit is better
            # the type-ignore for the strict=False is necessary for Python<3.10
            if level_self != level_other:
                return False
        # the iteration stopped meaning that for all levels that both other and self share, they are identical.
        # That's the definition of a sub stack. It also means that any stack is a sub stack of itself.
        return True

    def is_parent_of(self, other: "EdifactStack") -> bool:
        """
        Returns true iff this other stack is a sub stack of self.
        """
        return other.is_sub_stack_of(self)

    def to_json_path(self) -> str:
        """
        Transforms this instance into a JSON Path.
        """
        result: str = "$"
        # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
        # pylint: disable=not-an-iterable
        for level in self.levels:
            result += '["' + level.name + '"]'
            if level.index is not None:
                result += f"[{level.index}]"
            elif level.is_groupable:
                result += "[0]"
        return result



"""
This module provides a class to collect information which of need for all parsing functions
"""

from docx.table import Table
from pydantic import BaseModel, Field

from kohlrahbi.enums import RowType
from kohlrahbi.table_header import PruefiMetaData, TableHeader, get_tabstop_positions


# pylint: disable=too-few-public-methods
class Seed(BaseModel):
    """
    helper class to store all values to extract the AHB and the final AHB as dataframe
    """

    pruefidentifikatoren: list[str] = Field(default_factory=list)
    column_headers: list[str] = Field(default_factory=list)
    edifact_struktur_left_indent_position: int = 0
    middle_cell_left_indent_position: int = 0
    tabstop_positions: list[int] = Field(default_factory=list)
    last_two_row_types: list[RowType] = Field(default_factory=list)
    metadata: list[PruefiMetaData] = Field(default_factory=list)

    # why this classmethod?
    # to decouple the data structure of Elixir from the input data
    # more information can be found on https://www.attrs.org/en/stable/init.html#initialization
    @classmethod
    def from_table(cls, docx_table: Table) -> "Seed":
        """Prepare DataFrame for a new table with new Prüfidentifikatoren

        Args:
            item (Union[Paragraph, Table]): A paragraph or table from the docx
        """

        # the header cell with all pruefi information is the last cell in the first row
        # the first row contains the 'EDIFACT Struktur' column, which is not needed
        # often there is a second row with the pruefidentifikatoren information but it is not reliable for all tables
        # therefore we use the last cell in the first row which seems to be the most reliable
        header_cell_with_all_pruefi_information = docx_table.row_cells(0)[-1]

        table_header = TableHeader.from_header_cell(row_cell=header_cell_with_all_pruefi_information)

        pruefidentifikatoren = table_header.get_pruefidentifikatoren()

        # edifact struktur cell
        edifact_struktur_indicator_paragraph = docx_table.cell(row_idx=4, col_idx=0).paragraphs[0]
        edifact_struktur_left_indent_position = edifact_struktur_indicator_paragraph.paragraph_format.left_indent

        # middle cell
        middle_cell_indicator_paragraph = docx_table.cell(row_idx=4, col_idx=1).paragraphs[0]
        middle_cell_left_indent_position = middle_cell_indicator_paragraph.paragraph_format.left_indent
        tabstop_positions = get_tabstop_positions(middle_cell_indicator_paragraph)

        # metadata
        metadata = table_header.pruefi_meta_data

        base_column_names: list[str] = [
            "Segment Gruppe",
            "Segment",
            "Segment ID",
            "Datenelement",
            "Codes und Qualifier",
            "Beschreibung",
        ]
        columns = base_column_names + pruefidentifikatoren + ["Bedingung"]

        # Initialize help variables
        last_two_row_types: list[RowType] = [RowType.EMPTY, RowType.EMPTY]

        return cls(
            pruefidentifikatoren=pruefidentifikatoren,
            column_headers=columns,
            edifact_struktur_left_indent_position=edifact_struktur_left_indent_position,
            middle_cell_left_indent_position=middle_cell_left_indent_position,
            tabstop_positions=tabstop_positions,
            last_two_row_types=last_two_row_types,
            metadata=metadata,
        )



"""
This module provides the AhbTable class
"""

from pathlib import Path
from typing import Union

import pandas as pd
from efoli import get_format_of_pruefidentifikator
from more_itertools import peekable
from pydantic import BaseModel, ConfigDict, Field

from kohlrahbi.ahbtable.ahbsubtable import AhbSubTable
from kohlrahbi.logger import logger
from kohlrahbi.table_header import PruefiMetaData

_column_letter_width_mapping: dict[str, Union[float, int]] = {
    "A": 3.5,
    "B": 47,
    "C": 9,
    "D": 14,
    "E": 39,
    "F": 33,
    "G": 18,
    "H": 102,
}


class AhbTable(BaseModel):
    """
    This class contains the AHB table as you see it in the AHB documents, but in a machine readable format.
    """

    table: pd.DataFrame
    metadata: list[PruefiMetaData] = Field(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def fill_segment_gruppe_segment_dataelement(self) -> None:
        """
        For easier readability this functions adds the segment

        before
        ======

        Nachrichten-Kopfsegment
                                    UNH
                                    UNH    0062

        after
        =====

        Nachrichten-Kopfsegment
        Nachrichten-Kopfsegment     UNH
        Nachrichten-Kopfsegment     UNH    0062
        """

        latest_segement_gruppe: str = ""
        latest_segement: str = ""
        latest_datenelement: str = ""

        for _, row in self.table.iterrows():
            if row["Segment Gruppe"] != "":
                latest_segement_gruppe = row["Segment Gruppe"]

            if row["Segment"] != "":
                latest_segement = row["Segment"]

            if row["Datenelement"] != "":
                latest_datenelement = row["Datenelement"]

            if row["Segment Gruppe"] == "" and row["Codes und Qualifier"] != "" or row["Segment"] != "":
                row["Segment Gruppe"] = latest_segement_gruppe
                row["Segment"] = latest_segement
                row["Datenelement"] = latest_datenelement

    @classmethod
    def from_ahb_sub_table(cls, ahb_sub_table: AhbSubTable) -> "AhbTable":
        """
        Create an AHB table from an AHB sub table
        """
        return cls(table=ahb_sub_table.table, metadata=ahb_sub_table.table_meta_data.metadata)

    def append_ahb_sub_table(self, ahb_sub_table: AhbSubTable) -> None:
        """
        Append an AHB sub table to this AHB table instance
        """
        if self.table is None:
            self.table = ahb_sub_table.table
        else:
            self.table = pd.concat([self.table, ahb_sub_table.table], ignore_index=True)

    @staticmethod
    def line_contains_only_segment_gruppe(raw_line: pd.Series) -> bool:  # type:ignore[type-arg]
        """
        Returns true if the given raw line only contains some meaningful data in the "Segment Gruppe" key
        """
        for key, value in raw_line.items():
            if key == "Segment Gruppe":
                continue
            if value is not None and len(value.strip()) > 0:
                return False
        return True

    def sanitize(self) -> None:
        """
        In some cases there is the content of one cell splits in two.
        We need to merge the content into one cell and delete the deprecated cell afterwards.
        """
        index_of_lines_to_drop: list[int] = []

        iterable_ahb_table = peekable(self.table.iterrows())
        self.table.reset_index(drop=True, inplace=True)
        for _, row in iterable_ahb_table:
            # pylint: disable=unpacking-non-sequence # it is a tuple indeed
            index_of_next_row, next_row = iterable_ahb_table.peek(
                (
                    0,
                    pd.Series(
                        {
                            "Segment Gruppe": "",
                            "Segment": "",
                            "Codes und Qualifier": "",
                            "Beschreibung": "",
                            "Bedingung": "",
                        }
                    ),
                )
            )

            segment_gruppe_contains_multiple_lines = (
                "Segment Gruppe" in row
                and row["Segment Gruppe"]
                and AhbTable.line_contains_only_segment_gruppe(row)
                and not next_row["Segment Gruppe"].startswith("SG")
                and not next_row["Segment"]
            )
            if segment_gruppe_contains_multiple_lines:
                merged_segment_gruppe_content = " ".join([row["Segment Gruppe"], next_row["Segment Gruppe"]])
                row["Segment Gruppe"] = merged_segment_gruppe_content.strip()

                if not isinstance(index_of_next_row, int):
                    raise TypeError(
                        f"The 'index_of_next_row' must by of type `int` but it is '{type(index_of_next_row)}'"
                    )

                if index_of_next_row == 0:
                    # this case is only for the first and last row. These lines should not get deleted.
                    continue
                index_of_lines_to_drop.append(index_of_next_row)

        self.table.drop(index_of_lines_to_drop, inplace=True)
        self.table.reset_index(drop=True, inplace=True)

    def to_csv(self, pruefi: str, path_to_output_directory: Path) -> None:
        """
        Dump an AHB table of a given pruefi into a csv file.
        The csv file will be saved in the following directory structure:
            <path_to_output_directory>/csv/<edifact_format>/<pruefi>.csv
        """
        edifact_format = get_format_of_pruefidentifikator(pruefi)
        if edifact_format is None:
            logger.warning("'%s' is not a pruefidentifikator", pruefi)
            return

        csv_output_directory_path = path_to_output_directory / "csv" / str(edifact_format)
        csv_output_directory_path.mkdir(parents=True, exist_ok=True)

        self.fill_segment_gruppe_segment_dataelement()

        columns_to_export = list(self.table.columns)[:5] + [pruefi]
        columns_to_export.append("Bedingung")
        df_to_export = self.table[columns_to_export]

        df_to_export.to_csv(csv_output_directory_path / f"{pruefi}.csv", encoding="utf-8")
        logger.info("The csv file for %s is saved at %s", pruefi, csv_output_directory_path / f"{pruefi}.csv")

    # pylint: disable=too-many-locals
    def to_xlsx(self, pruefi: str, path_to_output_directory: Path) -> None:
        """
        Dump a AHB table of a given pruefi into an excel file.
        The excel file will be saved in the following directory structure:
            <path_to_output_directory>/xlsx/<edifact_format>/<pruefi>.xlsx
        """
        edifact_format = get_format_of_pruefidentifikator(pruefi)
        if edifact_format is None:
            logger.warning("'%s' is not a pruefidentifikator", pruefi)
            return

        xlsx_output_directory_path: Path = path_to_output_directory / "xlsx" / str(edifact_format)
        xlsx_output_directory_path.mkdir(parents=True, exist_ok=True)

        excel_file_name = f"{pruefi}.xlsx"

        columns_to_export = list(self.table.columns)[:5] + [pruefi]
        columns_to_export.append("Bedingung")
        df_to_export = self.table[columns_to_export]

        try:
            # https://github.com/PyCQA/pylint/issues/3060 pylint: disable=abstract-class-instantiated
            with pd.ExcelWriter(xlsx_output_directory_path / excel_file_name, engine="xlsxwriter") as writer:
                df_to_export.to_excel(writer, sheet_name=f"{pruefi}")
                # pylint: disable=no-member
                workbook = writer.book
                worksheet = writer.sheets[f"{pruefi}"]
                wrap_format = workbook.add_format({"text_wrap": True})
                for column_letter, column_width in _column_letter_width_mapping.items():
                    excel_header = f"{column_letter}:{column_letter}"
                    worksheet.set_column(excel_header, column_width, wrap_format)
                logger.info("💾 Saved file(s) for Pruefidentifikator %s", pruefi)
        except PermissionError:
            logger.error("The Excel file %s is open. Please close this file and try again.", excel_file_name)



"""
This module contains the function to write the collected conditions to a json file.
"""

from pathlib import Path

import docx
from efoli import EdifactFormat, EdifactFormatVersion, get_format_of_pruefidentifikator

from kohlrahbi.ahb import get_pruefi_to_file_mapping
from kohlrahbi.ahbtable.ahbcondtions import AhbConditions
from kohlrahbi.ahbtable.ahbpackagetable import AhbPackageTable
from kohlrahbi.conditions.allgemeine_festlegungen import time_conditions
from kohlrahbi.logger import logger
from kohlrahbi.read_functions import get_all_conditions_from_doc


def find_all_files_from_all_pruefis(pruefi_to_file_mapping: dict[str, str]) -> dict[EdifactFormat, list[str]]:
    """takes list of all pruefis with according files and returns a dict edifactformat-> list(filepaths)"""
    format_to_files_mapping: dict[EdifactFormat, list[str]] = {}
    for pruefi, filename in pruefi_to_file_mapping.items():
        if not filename:
            raise ValueError(f"No file provided for pruefi {pruefi}")
        edifact_format = get_format_of_pruefidentifikator(pruefi)
        if format_to_files_mapping.get(edifact_format) is None:
            format_to_files_mapping[edifact_format] = [filename]
        elif filename not in format_to_files_mapping[edifact_format]:
            format_to_files_mapping[edifact_format].append(filename)
    return format_to_files_mapping


def scrape_conditions(
    basic_input_path: Path,
    output_path: Path,
    format_version: EdifactFormatVersion,
) -> None:
    """
    starts the scraping process for conditions of all formats
    """
    path_to_file = basic_input_path / Path("edi_energy_de") / Path(format_version.value)
    pruefi_to_file_mapping = get_pruefi_to_file_mapping(basic_input_path, format_version)

    collected_conditions: AhbConditions = AhbConditions()
    collected_packages: AhbPackageTable = AhbPackageTable()
    all_format_files = find_all_files_from_all_pruefis(pruefi_to_file_mapping)
    for edifact_format, files in all_format_files.items():
        for file in files:
            # pylint: disable=too-many-function-args
            path: Path = basic_input_path / path_to_file / Path(file)
            doc = docx.Document(str(path.absolute()))
            logger.info("Start scraping conditions for %s in %s", edifact_format, file)
            if not doc:
                logger.error("Could not open file %s as docx", Path(file))
            packages, cond_table = get_all_conditions_from_doc(doc, edifact_format)
            if packages and packages.table is not None:
                collected_conditions.include_condition_dict(packages.provide_conditions(edifact_format))
                collected_packages.include_package_dict(packages.package_dict)
            collected_conditions.include_condition_dict(cond_table.conditions_dict)
        collected_conditions.include_condition_dict({edifact_format: time_conditions})
    collected_conditions.dump_as_json(output_path)
    collected_packages.dump_as_json(output_path)



"""
models are the data classes used to store information from e.g. ahbs.
This module contains methods available to all methods in the package.
"""


def _check_that_string_is_not_whitespace_or_empty(value: str) -> str:
    """
    Check that string value is not empty or consists only of whitespace.
    """
    if not value:
        raise ValueError("The string must not be None or empty")
    if len(value.strip()) == 0:
        raise ValueError(f"The string must not consist only of whitespace: '{value}'")
    return value



""" "
Command line interface for the pruefis module.
"""

import sys
from pathlib import Path

import click
from efoli import EdifactFormatVersion

from kohlrahbi.ahb import scrape_pruefis
from kohlrahbi.enums.ahbexportfileformat import AhbExportFileFormat


def check_python_version() -> None:
    """
    Check if the Python interpreter is greater or equal to 3.11
    """
    if sys.version_info.major != 3 or sys.version_info.minor < 11:
        raise click.Abort(
            f"""Python >=3.11 is required to run this script but you use Python
{sys.version_info.major}.{sys.version_info.minor}"""
        )


# pylint: disable=unused-argument
def validate_path(ctx, param, value) -> Path:  # type:ignore[no-untyped-def]
    """
    Ensure the path exists or offer to create it.
    """
    path = Path(value)
    if not path.exists():
        if ctx.params.get("assume_yes") or click.confirm(
            f"The path {value} does not exist. Would you like to create it?"
        ):
            path.mkdir(parents=True)
            click.secho(f"Created directory {path}.", fg="green")
        else:
            click.secho("👋 Alright I will end this program now. Have a nice day.", fg="green")
            raise click.Abort()
    return path


@click.command()
@click.option(
    "-p",
    "--pruefis",
    default=[],
    required=False,
    help="Five digit number like 11042 or use wildcards like 110* or *042 or 11?42.",
    multiple=True,
)
@click.option(
    "-eemp",
    "--edi-energy-mirror-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path),
    help="The root path to the edi_energy_mirror repository.",
    required=True,
)
@click.option(
    "-o",
    "--output-path",
    type=click.Path(exists=False, dir_okay=True, file_okay=False, resolve_path=True, path_type=Path),
    callback=validate_path,
    default="output",
    prompt="Output directory",
    help="""Define the path where you want to save the generated files. If the path does not exist,
you will be asked if you want to create it.""",
)
@click.option(
    "--file-type",
    type=click.Choice([aeff.value for aeff in AhbExportFileFormat], case_sensitive=False),
    multiple=True,
    required=True,
    help="File type(s) for the scraped AHB documents.",
)
@click.option(
    "--format-version",
    multiple=False,
    type=click.Choice([e.value for e in EdifactFormatVersion], case_sensitive=False),
    required=True,
    help="Format version(s) of the AHB documents, e.g. FV2310",
)
@click.option(
    "--assume-yes",
    "-y",
    is_flag=True,
    default=False,
    help="Confirm all prompts automatically.",
)
@click.option(
    "--clear-output-path",
    is_flag=True,
    default=False,
    help="Clear old removed files from existing output path",
)
# pylint: disable=too-many-arguments, too-many-positional-arguments
def ahb(
    pruefis: list[str],
    edi_energy_mirror_path: Path,
    output_path: Path,
    file_type: tuple[AhbExportFileFormat, ...],
    format_version: EdifactFormatVersion | str,
    assume_yes: bool,  # pylint: disable=unused-argument
    clear_output_path: bool = False,
    # it is used by the callback function of the output-path
) -> None:
    """
    Scrape AHB documents for pruefidentifikatoren.
    This is a command line interface for the pruefis module.
    """
    check_python_version()
    if isinstance(format_version, str):
        format_version = EdifactFormatVersion(format_version)

    scrape_pruefis(
        pruefis=list(pruefis),
        basic_input_path=edi_energy_mirror_path,
        output_path=output_path,
        file_type=file_type,
        format_version=format_version,
        clear_output_path=clear_output_path,
    )



"""
This module contains the RowType enumeration.
"""

from enum import StrEnum, auto


class RowType(StrEnum):
    """
    All possible row types.
    The RowType is defined by the first cell in each row.
    Example content for each row type is documented at each enum member.
    """

    HEADER = auto()  #: e.g. "EDIFACT Struktur"
    SEGMENTNAME = auto()  #: e.g. "Nachrichten-Kopfsegment"
    SEGMENTGRUPPE = auto()  #: e.g. "SG2"
    SEGMENT = auto()  #: e.g. "    UNH" or "SG2  NAD"
    DATENELEMENT = auto()  #: e.g. "    UNH 0062"
    EMPTY = auto()  #: e.g. ""



"""
This module contains the UnfoldedAhbLine class.
"""

from pydantic import BaseModel


# pylint: disable=too-few-public-methods
class UnfoldedAhbLine(BaseModel):
    """
    This class represents one unfolded line of the AHB.
    Unfolded means that we separate segment_name and segment_gruppe as well as code and qualifier

    Example:

    AhbLine(
        guid=None,
        section_name="Ansprechpartner",
        segment_group_key="SG3",
        segment_code=None,
        data_element=None,
        value_pool_entry=None,
        name=None,
        ahb_expression="Kann",
        index=0,
    )

    AhbLine(
        guid=None,
        section_name="Ansprechpartner",
        segment_group_key="SG3",
        segment_code=None,
        data_element=None,
        value_pool_entry=None,
        name=None,
        ahb_expression="Kann",
        index=1,
    )

    AhbLine(
        guid=None,
        section_name="Ansprechpartner",
        segment_group_key="SG3",
        segment_code="CTA",
        data_element=None,
        value_pool_entry=None,
        name=None,
        ahb_expression="Muss",
        index=2,
    )

    AhbLine(
        guid=None,
        section_name="Ansprechpartner",
        segment_group_key="SG3",
        segment_code="CTA",
        data_element="3139",
        value_pool_entry="IC",
        name="Informationskontakt",
        ahb_expression="X",
        index=3,
    )

    AhbLine(
        guid=None,
        section_name="Ansprechpartner",
        segment_group_key="SG3",
        segment_code="CTA",
        data_element="3412",
        value_pool_entry=None,
        name="Name vom Ansprechpartner",
        ahb_expression="X",
        index=4,
    )

    """

    # pylint: disable=fixme
    # TODO: add class variable which increments the index with each created instance
    index: int
    segment_name: str  # Ansprechpartner
    segment_gruppe: str | None  # SG3
    segment: str | None  # CTA
    segment_id: str | None = None  # 00009
    datenelement: str | None  # 3055
    code: str | None  # IC
    qualifier: str | None  # Name vom Ansprechpartner
    beschreibung: str | None  # Informationskontakt
    bedingung_ausdruck: str | None  # Muss / X / X [1P0..1]
    bedingung: str | None  # [492] Wenn MP-ID in NAD+MR (Nachrichtenempfänger) aus Sparte Strom



"""
This module contains the UnfoldedAhbTableMetaData class.
"""

from pydantic import BaseModel


# pylint: disable=too-few-public-methods
class UnfoldedAhbTableMetaData(BaseModel):
    """
    This class represents the metadata of an unfolded AHB table.
    """

    pruefidentifikator: str
    beschreibung: str | None
    kommunikation_von: str | None



"""
This module contains the enumeration for FlatAHB row types.
"""

from enum import StrEnum, auto


class FlatAhbRowType(StrEnum):
    """
    All possible row types.
    The RowType is defined by the first cell in each row.
    Example content for each row type is documented at each enum member.
    """

    SEGMENTOPENINGLINE = auto()  #: e.g. "Nachrichten-Kopfsegment"
    # SECTIONNAME:



"""
A module that reads AHBs from a given CSV file
"""

import csv
import logging
import re
import uuid
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Literal, Optional, Sequence, TextIO, Tuple, overload

from kohlrahbi.models.anwendungshandbuch import _VERSION, AhbLine, AhbMetaInformation, FlatAnwendungshandbuch
from kohlrahbi.models.edifact_components import gabi_edifact_qualifier_pattern

_pruefi_pattern = re.compile(r"^\d{5}$")  #: five digits
_value_pool_entry_pattern = re.compile(r"^(?!MP-ID)[A-Z0-9\-i]{2,}$")  # i for GABi -- why?
_numeric_value_pool_entry_pattern = re.compile(r"^\d+(?:\.\d+)?[a-z]?$")
_ebd_code_pattern = re.compile(r"^([EG])_\d+$")
_segment_group_pattern = re.compile(r"^SG\d+$")


# pylint:disable=too-few-public-methods
class FlatAhbReader(ABC):
    """
    An AHB Reader reads AHB data from a source and is able to convert them to a :class:`.FlatAnwendungshandbuch`
    """

    @abstractmethod
    def to_flat_ahb(self) -> FlatAnwendungshandbuch:
        """
        convert the read data into a flat anwendungshandbuch
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")


def check_file_can_be_parsed_as_ahb_csv(file_path: Path) -> None:
    """
    Returns nothing iff the given file is parsable as CSV and contains no obvious errors.
    This is not a really sophisticated analysis but just a basic minimal sanity check.
    In case of error an exception is raised.
    """
    _ = FlatAhbCsvReader(file_path)  # this may die with a meaningful exception


class FlatAhbCsvReader(FlatAhbReader):
    """
    reads csv files and returns AHBs
    """

    def __init__(  # type:ignore[no-untyped-def]
        self,
        file_path: Path,
        pruefidentifikator: Optional[str] = None,
        encoding="utf-8",
        delimiter=",",
    ):
        self.rows: list[AhbLine] = []
        self._logger = logging.getLogger()
        self.current_section_name: Optional[str] = None
        self.pruefidentifikator = pruefidentifikator
        self.delimiter = delimiter
        self.bedingungen: dict[str, str] = {}
        with open(file_path, "r", encoding=encoding) as infile:
            # current_section_name: Optional[str]
            raw_lines = self.get_raw_rows(infile)
        raw_lines_with_merged_section_names = FlatAhbCsvReader.merge_section_only_lines(raw_lines)
        for row_index, row in enumerate(raw_lines_with_merged_section_names):
            ahb_line = self.raw_ahb_row_to_ahbline(row)
            if ahb_line is None:
                continue
            ahb_line.index = row_index  # it is ascending but not continuous
            self.rows.append(ahb_line)

    @staticmethod
    def merge_section_only_lines(raw_lines: list[dict]) -> list[dict]:  # type:ignore[type-arg]
        """
        merges adjacent lines from the CSV source when they only contain an AHB "section" description.
        "Section" headings are the grey lines on the left of the AHB PDF.
        (The first section of each AHB is "Nachrichten-Kopfsegment" in most cases.)
        When the section heading spans multiple lines, we don't want to treat them as separate but as a single heading.
        The method consumes a list of dicts and returns a _new_ list of dicts that is of the same length or shorter.
        """
        result: list[dict] = []  # type:ignore[type-arg]

        # imagine the original list to be
        # 0,asd,qwertz,
        # 1,a very long section,
        # 2,heading that spans,
        # 3,multiple lines,
        # 4,Foo,Bar,Y
        # 5,Baz,Boom,Z
        # we then want to merge the lines with index 1-3 into a single line
        keys_that_must_no_hold_any_values: set[str] = {
            "Segment",
            "Datenelement",
            "Codes und Qualifier",
            "Beschreibung",
            "Bedingung",
        }

        def line_only_contains_segment_gruppe(raw_line: dict) -> bool:  # type:ignore[type-arg]
            """
            returns true if the given raw line only contains some meaningful data in the "Segment Gruppe" key
            """
            for row_key in keys_that_must_no_hold_any_values:
                if row_key in raw_line and raw_line[row_key] is not None and len(raw_line[row_key].strip()) > 0:
                    return False
            return True

        merged_section_name = ""
        number_of_lines_merged = 0
        for raw_line in raw_lines:
            if (
                "Segment Gruppe" in raw_line
                and raw_line["Segment Gruppe"]
                and line_only_contains_segment_gruppe(raw_line)
                and not raw_line["Segment Gruppe"].startswith("SG")
            ):
                merged_section_name += " " + raw_line["Segment Gruppe"]
                number_of_lines_merged += 1
            else:
                # note that AHBs never end with a section heading, so all headings/sections will run into this block
                if len(merged_section_name) > 0:
                    artificial_merged_line: dict = {  # type:ignore[type-arg]
                        "": str(int(raw_line[""]) - 1),
                        "Segment Gruppe": merged_section_name.strip().replace("  ", " "),
                    }
                    for key in keys_that_must_no_hold_any_values:
                        # although we know there's no meaningful value here, we still need the keys with empty values
                        # so that to downstream code the line seems legit ➡ We re-add them.
                        artificial_merged_line[key] = ""
                    result.append(artificial_merged_line)
                    merged_section_name = ""
                    number_of_lines_merged = 0
                result.append(raw_line)
        return result

    def get_raw_rows(self, file_handle: TextIO) -> list[dict]:  # type:ignore[type-arg]
        """
        reads the input file and returns an iterator over raw lines.
        Override this method if your data source is not a CSV file
        """
        reader = csv.DictReader(file_handle, delimiter=self.delimiter)
        if not self.pruefidentifikator:
            self.pruefidentifikator = FlatAhbCsvReader._get_name_of_expression_column(reader.fieldnames)
        if not self.pruefidentifikator:
            raise ValueError("Cannot find column name for ahb expression")
        return list(reader)

    def raw_ahb_row_to_ahbline(self, ahb_row: dict) -> Optional[AhbLine]:  # type:ignore[type-arg]
        """
        Converts a row of the raw/scraped AHB into the AhbLine data structure.
        Returns None for rows that are skipped.
        Override this method is your column names/input data structure differs.
        """
        value_pool_entry, description = FlatAhbCsvReader.separate_value_pool_entry_and_name(
            ahb_row["Codes und Qualifier"], ahb_row["Beschreibung"]
        )
        self.bedingungen.update(FlatAhbCsvReader._extract_bedingungen(ahb_row["Bedingung"]))
        segment_group: Optional[str] = None
        if FlatAhbCsvReader._is_segment_group(ahb_row["Segment Gruppe"]):
            segment_group = ahb_row["Segment Gruppe"]
        elif len(ahb_row["Segment Gruppe"]) >= 3:
            self.current_section_name = ahb_row["Segment Gruppe"].strip() or None  # e.g. "Nachrichten-Kopfsegment"
            self._logger.debug("Processing %s section '%s'", self.pruefidentifikator, self.current_section_name)
            return None  # possibly a section heading like "Nachrichten-Endesegment"
            # this is different from segment group = None which is value for e.g. the UNH
        result: AhbLine = AhbLine(
            guid=uuid.uuid4(),
            segment_group_key=segment_group,
            segment_code=ahb_row["Segment"] or None,
            data_element=ahb_row["Datenelement"] or None,
            value_pool_entry=value_pool_entry,
            ahb_expression=ahb_row[self.pruefidentifikator] or None,
            name=description,
            section_name=_replace_hardcoded_section_names(self.current_section_name),
        )
        return result

    def extract_condition_texts(self) -> dict[str, str]:
        """
        Extracts all the condition texts found in this AHB.
        :return: a dictionary with the condition key (e.g. "46") as value and the condition text (e.g. "Wenn aus Sparte
        Gas") as value. The value does not contain the "[46]" prefix.
        """
        return self.bedingungen

    @staticmethod
    def separate_value_pool_entry_and_name(
        x: Optional[str], y: Optional[str]  # pylint:disable=invalid-name
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        The PDFs are so broken, that sometimes the Codes column contains the description of the line instead of the code
        This function returns a value_pool_entry at index 0, a possible description at index 1, even if they're mixed up
        in the source files.
        """
        if x is not None:
            x = x.strip()
        if y is not None:
            y = y.strip()
        if FlatAhbCsvReader._is_value_pool_entry(x) and not FlatAhbCsvReader._is_value_pool_entry(y):
            return x, y or None
        if FlatAhbCsvReader._is_value_pool_entry(x) and FlatAhbCsvReader._is_value_pool_entry(y):
            # Both look like a value pool entry. This typically happens e.g. for date qualifiers or code lists
            return x, y
        return y or None, x or None

    @staticmethod
    def _is_value_pool_entry(candidate: Optional[str]) -> bool:
        """
        Returns true iff the provided candidate is a possible value pool entry.
        """
        if not candidate:
            return False
        if _value_pool_entry_pattern.match(candidate) is not None:
            return True
        # numbers alone might be value pool entries even if they don't match the regex
        # we don't use "isdigit" because isdigit e.g. does not match '1.2'
        if _numeric_value_pool_entry_pattern.match(candidate) is not None:
            return True
        if len(candidate) == 1 and candidate.upper() == candidate:
            return True
        if gabi_edifact_qualifier_pattern.match(candidate) is not None:
            return True
        return _ebd_code_pattern.match(candidate) is not None

    @staticmethod
    def _is_segment_group(candidate: Optional[str]) -> bool:
        """
        Returns true iff the provided candidate is a possible segment group
        """
        if not candidate:
            return False
        return _value_pool_entry_pattern.match(candidate) is not None

    _bedingung_pattern = re.compile(r"\[(?P<key>\d+)\]\s*(?P<text>[^\[\]]+)\s*")  # https://regex101.com/r/hN5x9w/1

    @staticmethod
    def _extract_bedingungen(candidate: Optional[str]) -> dict[str, str]:
        """
        Checks if the given candidate is a bedingung. If no, returns empty dict.
        If yes returns a dictionary where the Bedingung keys (e.g. "494") are dictionary keys and the Bedingung
        texts (e.g. "Der Wert muss ≤ der zum Erzeugungszeitpunkt sein.") are the dictionary values.
        The result is a dict because often there are multiple conditions connected to a single line in the AHB.
        """
        if not candidate:
            return {}
        return {m[0]: m[1].strip() for m in FlatAhbCsvReader._bedingung_pattern.findall(candidate)}

    @staticmethod
    def _get_name_of_expression_column(field_names: Optional[Sequence[str]]) -> Optional[str]:
        """
        Gets the name of the column that holds the AHB expressions.
        This allows us to read any AHB without a priori knowledge of its pruefidentifikator.

        :param field_names: list of fieldnames
        :return: the first field name that is a 5 digit name or None if none is found.
        """
        if not field_names:
            return None
        for field_name in field_names:
            if _pruefi_pattern.match(field_name):
                return field_name
        return None

    def to_flat_ahb(self) -> FlatAnwendungshandbuch:
        """
        Converts the content of the CSV file to a FlatAnwendungshandbuch.
        :return:
        """
        return FlatAnwendungshandbuch(
            meta=AhbMetaInformation(
                pruefidentifikator=self.pruefidentifikator,  # type:ignore[arg-type]
                maus_version=_VERSION,
            ),
            lines=[row for row in self.rows if row.holds_any_information()],
        )


@overload
def _replace_hardcoded_section_names(section_name: str) -> str: ...


@overload
def _replace_hardcoded_section_names(section_name: Literal[None]) -> Literal[None]: ...


def _replace_hardcoded_section_names(section_name: Optional[str]) -> Optional[str]:
    """
    Replace section names that differ because of "Datenschiefstände"
    :param section_name:
    :return:
    """
    if not section_name:
        return section_name
    replacements: dict[str, str] = {
        # https://github.com/Hochfrequenz/edifact-templates/issues/82
        # pylint: disable=line-too-long
        "OBIS-Kennzahl der Zähleinrichtung / Mengenumwerter / Smartmeter-Gateway": "OBIS-Kennzahl der Zähleinrichtung / Mengenumwerter",
        "OBIS-Daten der Zähleinrichtung / Mengenumwerter / Smartmeter-Gateway": "OBIS-Daten der Zähleinrichtung / Mengenumwerter",
        # https://github.com/Hochfrequenz/edifact-templates/issues/80
        "OBIS Daten für Lieferant relevant": "OBIS Daten für Marktrolle relevant",
    }
    if section_name.strip() in replacements:
        return replacements[section_name]
    return section_name



"""
kohlrahbi is a package to scrape AHBs (in docx format)
"""

import click

from kohlrahbi.ahb.command import ahb
from kohlrahbi.changehistory.command import changehistory
from kohlrahbi.conditions.command import conditions
from kohlrahbi.qualitymap.command import qualitymap
from kohlrahbi.version import version


@click.group()
@click.version_option(version=version)
def cli() -> None:
    """Kohlrahbi CLI tool"""


# Add commands to the CLI group
cli.add_command(ahb)
cli.add_command(changehistory)
cli.add_command(conditions)
cli.add_command(qualitymap)

if __name__ == "__main__":
    # the parameter arguments gets provided over the CLI
    cli()  # pylint:disable=no-value-for-parameter



"""
This module provides functions for scraping change histories from .docx files and saving them to an Excel file.

The main functions in this module are:
- `scrape_change_histories`: Starts the scraping process of the change histories.
- `find_docx_files`: Finds all .docx files containing change histories.
- `process_docx_file`: Reads and processes change history from a .docx file.
- `save_change_histories_to_excel`: Saves the collected change histories to an Excel file.
- `create_sheet_name`: Creates a sheet name from the filename.
"""

from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

import docx
import pandas as pd
from docx.document import Document
from docx.table import Table

from kohlrahbi.changehistory.changehistorytable import ChangeHistoryTable
from kohlrahbi.docxfilefinder import DocxFileFinder
from kohlrahbi.logger import logger
from kohlrahbi.read_functions import get_all_paragraphs_and_tables


def is_change_history_table(table: Table) -> bool:
    """
    Checks if the given table is change history table.
    """
    # in the document 'Entscheidungsbaum-DiagrammeundCodelisten-informatorischeLesefassung3.5_99991231_20240401.docx'
    # I got the error "IndexError: list index out of range", I am not sure which table caused the error
    try:
        return table.cell(row_idx=0, col_idx=0).text.strip() == "Änd-ID"
    except IndexError:
        return False


def get_change_history_table(document: Document) -> Optional[ChangeHistoryTable]:
    """
    Reads a docx file and extracts the change history.
    Returns None if no such table was found.
    """

    # Iterate through the whole word document
    logger.info("🔁 Start iterating through paragraphs and tables")
    for item in get_all_paragraphs_and_tables(parent=document):
        if isinstance(item, Table) and is_change_history_table(table=item):
            change_history_table = ChangeHistoryTable.from_docx_change_history_table(docx_table=item)
            return change_history_table

    return None


def save_change_histories_to_excel(change_history_collection: dict[str, pd.DataFrame], output_path: Path) -> None:
    """
    Save the collected change histories to an Excel file.
    """
    # add timestamp to file name
    # there are two timestamps: one with datetime and another one with just date information.
    # It is handy during debugging to save different versions of the output files with the datetime information.
    # But in production we only want to save one file per day.
    # current_timestamp = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
    current_timestamp = datetime.now(tz=timezone.utc).strftime("%Y-%m-%d")
    path_to_change_history_excel_file = output_path / f"{current_timestamp}_change_histories.xlsx"

    logger.info("💾 Saving change histories xlsx file %s", path_to_change_history_excel_file)

    # Define column widths (example: 20, 15, 30, etc.)
    column_widths = [6, 6, 46, 52, 52, 38, 15]  # Replace with your desired widths

    # Create a Pandas Excel writer using XlsxWriter as the engine
    # https://github.com/PyCQA/pylint/issues/3060 pylint: disable=abstract-class-instantiated
    with pd.ExcelWriter(path_to_change_history_excel_file, engine="xlsxwriter") as writer:
        for sheet_name, df in change_history_collection.items():
            df.to_excel(writer, sheet_name=sheet_name)

            # Access the XlsxWriter workbook and worksheet objects
            workbook = writer.book
            worksheet = writer.sheets[sheet_name]

            # Create a text wrap format, this is needed to avoid the text being cut off in the cells
            wrap_format = workbook.add_format({"text_wrap": True})

            # Get the dimensions of the DataFrame
            (_, max_col) = df.shape

            assert max_col + 1 == len(column_widths)  # +1 cause of index

            # Apply text wrap format to each cell
            for col_num, width in enumerate(column_widths):
                worksheet.set_column(col_num, col_num, width, wrap_format)


def create_sheet_name(filename: str) -> str:
    """
    Creates a sheet name from the filename.

    We need to shorten the sheet name because Excel only allows 31 characters for sheet names.
    This function replaces some words with acronyms and removes some words.
    """
    sheet_name = filename.split("-informatorischeLesefassung")[0]

    if "Entscheidungsbaum-Diagramm" in sheet_name:
        sheet_name = sheet_name.replace("Entscheidungsbaum", "EBDs")
    if "Artikelnummern" in sheet_name:
        sheet_name = sheet_name.replace("Artikelnummern", "Artikelnr")
    if "Codeliste" in sheet_name:
        sheet_name = sheet_name.replace("Codeliste", "CL")
    if len(sheet_name) > 31:
        # Excel only allows 31 characters for sheet names
        # but REQOTEQUOTESORDERSORDRSPORDCHGAHB is 33 characters long
        sheet_name = sheet_name.replace("HG", "")
    return sheet_name


def find_docx_files(input_path: Path) -> list[Path]:
    """
    Find all .docx files containing change histories.
    """
    docx_file_finder = DocxFileFinder.from_input_path(input_path=input_path)
    return docx_file_finder.get_all_docx_files_which_contain_change_histories()


def process_docx_file(file_path: Path) -> Optional[pd.DataFrame]:
    """
    Read and process change history from a .docx file.
    """
    doc = docx.Document(str(file_path))
    logger.info("🤓 Start reading docx file '%s'", str(file_path))
    change_history_table = get_change_history_table(document=doc)

    if change_history_table is not None:
        change_history_table.sanitize_table()
        return change_history_table.table
    return None


def scrape_change_histories(input_path: Path, output_path: Path) -> None:
    """
    starts the scraping process of the change histories
    """
    logger.info("👀 Start looking for change histories")
    ahb_file_paths = find_docx_files(input_path)

    change_history_collection = {}
    for file_path in ahb_file_paths:
        df = process_docx_file(file_path)
        if df is not None:
            change_history_collection[create_sheet_name(file_path.name)] = df

    save_change_histories_to_excel(change_history_collection, output_path)



"""
This module provides a logger instance for kohlrahbi
"""

import logging.config
from pathlib import Path

# Construct the path to the configuration file
logger_config_file: Path = Path(__file__).with_suffix(".ini")

logging.config.fileConfig(logger_config_file)
logger = logging.getLogger("kohlrahbi")



"""
This module contains the class BodyCell
"""

import pandas as pd
from docx.table import _Cell
from docx.text.paragraph import Paragraph
from pydantic import BaseModel, ConfigDict

from kohlrahbi.models.flat_ahb_reader import FlatAhbCsvReader
from kohlrahbi.table_header import get_tabstop_positions

INDEX_OF_CODES_AND_QUALIFIER_COLUMN = 4
KNOW_SUFFIXES = {
    "g",
    "ung",
    "gs-",
    "vall",
    "n",
    "m",
    "t",
    "rage",
    "sgrund",
}  # only a temporary and incomplete list to  filter some cases, not intended as NLP


class BodyCell(BaseModel):
    """
    BodyCell contains all information and a method
    to extract dataelement/qualifier, the name of the dataelement
    as well as the conditions for each Prüfidentifikator.
    """

    table_cell: _Cell
    left_indent_position: int
    indicator_tabstop_positions: list[int]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # I see why pylint is not happy about this many branches, but at the moment I have no clue how to avoid them.
    # pylint: disable=too-many-branches
    def parse(self, ahb_row_dataframe: pd.DataFrame) -> pd.DataFrame:
        """Parses a paragraph in the middle column and puts the information into the appropriate columns

        Args:
            paragraph (Paragraph): Current paragraph in the edifact struktur cell
            dataframe (pd.DataFrame): Contains all infos
            left_indent_position (int): Position of the left indent from the indicator middle cell
            tabstop_positions (list[int]): All tabstop positions of the indicator middle cell
        """

        def add_text_to_column(row_index: int, column_index: int, text: str) -> None:
            starts_with_known_suffix = any(text.startswith(suffix + " ") for suffix in KNOW_SUFFIXES)
            if len(text) > 0:
                if (
                    len(ahb_row_dataframe.iat[row_index, column_index]) > 0
                    and not starts_with_known_suffix
                    and len(text) > 1
                ):
                    text = " " + text
                ahb_row_dataframe.iat[row_index, column_index] += text

        def handle_code_or_qualifier_entry(
            splitted_text_at_tabs: list[str], row_index: int, is_first_iteration: bool
        ) -> int:
            if (  # pylint: disable=protected-access
                FlatAhbCsvReader._is_value_pool_entry(candidate=splitted_text_at_tabs[0])
                and len(splitted_text_at_tabs) >= 2
            ):
                if not is_first_iteration:
                    ahb_row_dataframe.loc[ahb_row_dataframe.index.max() + 1, :] = ""
                    row_index += 1
            add_text_to_column(row_index, INDEX_OF_CODES_AND_QUALIFIER_COLUMN, splitted_text_at_tabs.pop(0))
            return row_index

        def handle_tab_stops(
            paragraph: Paragraph, splitted_text_at_tabs: list[str], row_index: int, column_indezes: list[int]
        ) -> None:
            tab_stops_in_current_paragraph = get_tabstop_positions(paragraph=paragraph)
            if len(tab_stops_in_current_paragraph) == len(splitted_text_at_tabs) - 1:
                # we have remaining parts from a qualifier or code
                tab_stops_in_current_paragraph = [
                    paragraph.paragraph_format.left_indent,
                    *tab_stops_in_current_paragraph,
                ]
            for tabstop in tab_stops_in_current_paragraph:
                for indicator_tabstop_position, column_index in zip(self.indicator_tabstop_positions, column_indezes):
                    if tabstop == indicator_tabstop_position:
                        add_text_to_column(row_index, column_index, splitted_text_at_tabs.pop(0))

        def handle_no_tab_stops(splitted_text_at_tabs: list[str], row_index: int) -> None:
            if splitted_text_at_tabs:
                column_index = ahb_row_dataframe.columns.get_loc("Beschreibung")
                assert isinstance(column_index, int)
                add_text_to_column(row_index, column_index, splitted_text_at_tabs.pop(0))

        cell_is_empty = self.table_cell.paragraphs[0].text == ""
        if cell_is_empty:
            return ahb_row_dataframe

        is_first_iteration = True

        for paragraph in self.table_cell.paragraphs:
            row_index = ahb_row_dataframe.index.max()
            paragraph.text = paragraph.text.replace("\xa0", "")
            splitted_text_at_tabs = paragraph.text.split("\t")

            if paragraph.paragraph_format.left_indent == self.left_indent_position:
                row_index = handle_code_or_qualifier_entry(splitted_text_at_tabs, row_index, is_first_iteration)
                column_indezes = list(
                    range(
                        INDEX_OF_CODES_AND_QUALIFIER_COLUMN + 1,
                        INDEX_OF_CODES_AND_QUALIFIER_COLUMN + 1 + len(self.indicator_tabstop_positions),
                    )
                )
            else:
                if splitted_text_at_tabs[0] == "":
                    del splitted_text_at_tabs[0]
                column_indezes = list(
                    range(
                        INDEX_OF_CODES_AND_QUALIFIER_COLUMN + 1,
                        INDEX_OF_CODES_AND_QUALIFIER_COLUMN + 1 + len(self.indicator_tabstop_positions),
                    )
                )

            if self.has_paragraph_tabstops(paragraph=paragraph):
                handle_tab_stops(paragraph, splitted_text_at_tabs, row_index, column_indezes)
            else:
                handle_no_tab_stops(splitted_text_at_tabs, row_index)

            is_first_iteration = False

        return ahb_row_dataframe

    def has_paragraph_tabstops(self, paragraph: Paragraph) -> bool:
        """
        Checks if the given paragraph contains tabstops
        """
        tab_stops = list(paragraph.paragraph_format.tab_stops)
        return any(tab_stops)



"""model classes copy-pasted from kohlrahbi"""

# pylint: disable=too-few-public-methods

try:
    from sqlalchemy import UniqueConstraint
    from sqlmodel import Field, Relationship, SQLModel
except ImportError as import_error:
    import_error.msg += "; Did you install kohlrahbi[sqlmodels]?"
    # sqlmodel is only an optional dependency when kohlrahbi is used to fill a database
    raise
import uuid
from typing import Optional
from uuid import UUID

from efoli import EdifactFormat, EdifactFormatVersion

# the models here do NOT inherit from the original models, because I didn't manage to fix the issue that arise:
# https://github.com/Hochfrequenz/kohlrahbi/issues/556
# <frozen abc>:106: in __new__
#     ???
# E   TypeError: FlatAnwendungshandbuch.__init_subclass__() takes no keyword arguments
#
# or
#
# ..\.tox\dev\Lib\site-packages\sqlmodel\main.py:697: in get_sqlalchemy_type
#     raise ValueError(f"{type_} has no matching SQLAlchemy type")
# E   ValueError: <class 'kohlrahbi.models.anwendungshandbuch.AhbMetaInformation'> has no matching SQLAlchemy type


class FlatAnwendungshandbuch(SQLModel, table=True):  # team
    """
    A flat Anwendungshandbuch (AHB) models an Anwendungshandbuch as combination of some meta data and an ordered list
    of `.class:`.FlatAhbLine s. Basically a flat Anwendungshandbuch is the result of a simple scraping approach.
    You can create instances of this class without knowing anything about the "under the hood" structure of AHBs or MIGs
    """

    id: UUID = Field(primary_key=True, default_factory=uuid.uuid4, description="optional key")
    meta: Optional["AhbMetaInformation"] = Relationship(back_populates="flatanwendungshandbuch")
    lines: list["AhbLine"] = Relationship(back_populates="flatanwendungshandbuch")


class AhbLine(SQLModel, table=True):
    """
    An AhbLine is a single line inside the machine-readable, flat AHB.
    """

    __table_args__ = (UniqueConstraint("ahb_id", "position_inside_ahb", name="IX_position_once_per_ahb"),)
    id: UUID = Field(primary_key=True, default_factory=uuid.uuid4, description="optional key")
    # yes, it's actually that bad already

    position_inside_ahb: int = Field(index=True)
    ahb_id: UUID | None = Field(default=None, foreign_key="flatanwendungshandbuch.id")
    flatanwendungshandbuch: FlatAnwendungshandbuch | None = Relationship(back_populates="lines")

    # fields copy-pasted from original model:
    guid: Optional[UUID]
    segment_group_key: Optional[str]
    segment_code: Optional[str]
    data_element: Optional[str]
    segment_id: Optional[str]
    value_pool_entry: Optional[str]
    name: Optional[str]
    ahb_expression: Optional[str]
    conditions: Optional[str]
    section_name: Optional[str]
    index: Optional[int]


class AhbMetaInformation(SQLModel, table=True):
    """
    Meta information about an AHB like e.g. its title, Prüfidentifikator, possible sender and receiver roles
    """

    __table_args__ = (
        UniqueConstraint("pruefidentifikator", "edifact_format_version", name="IX_pruefi_once_per_format_version"),
    )
    id: UUID = Field(primary_key=True, default_factory=uuid.uuid4, description="optional key")
    edifact_format: EdifactFormat = Field(index=True)
    edifact_format_version: EdifactFormatVersion = Field(index=True)
    flatanwendungshandbuch: Optional[FlatAnwendungshandbuch] = Relationship(
        back_populates="meta", sa_relationship_kwargs={"uselist": False}
    )
    ahb_id: UUID | None = Field(default=None, foreign_key="flatanwendungshandbuch.id")

    # copy-pasted fields from original model:
    pruefidentifikator: str
    maus_version: Optional[str]
    description: Optional[str]
    direction: Optional[str]



"""
This module contains the AhbSubTable class.
"""

from typing import Generator, Union

import numpy as np
import pandas as pd
from docx.table import Table as DocxTable
from docx.table import _Cell, _Row
from docx.text.paragraph import Paragraph
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict

from kohlrahbi.ahbtable.ahbtablerow import AhbTableRow
from kohlrahbi.docxtablecells.bodycell import INDEX_OF_CODES_AND_QUALIFIER_COLUMN, KNOW_SUFFIXES
from kohlrahbi.enums import RowType
from kohlrahbi.row_type_checker import get_row_type
from kohlrahbi.seed import Seed


class AhbSubTable(BaseModel):
    """
    The AHB table for one Pruefidentifikator is separated into small sub tables.
    This class contains the information from such a sub table.
    """

    table_meta_data: Seed
    table: pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @staticmethod
    def _parse_docx_table(
        table_meta_data: Seed, ahb_table_dataframe: pd.DataFrame, docx_table: DocxTable
    ) -> pd.DataFrame:
        """Parse the docx table and add the information to the dataframe."""
        for row in docx_table.rows:
            sanitized_cells = list(AhbSubTable.iter_visible_cells(row=row))

            current_edifact_struktur_cell = sanitized_cells[0]

            # check for row type
            current_row_type = get_row_type(
                edifact_struktur_cell=current_edifact_struktur_cell,
                left_indent_position=table_meta_data.edifact_struktur_left_indent_position,
            )

            edifact_struktur_cell = sanitized_cells[0]
            middle_cell = sanitized_cells[1]
            bedingung_cell = sanitized_cells[-1]

            # this case covers the "normal" docx table row
            if not (current_row_type is RowType.EMPTY and table_meta_data.last_two_row_types[0] is RowType.HEADER):
                ahb_table_row = AhbTableRow(
                    seed=table_meta_data,
                    edifact_struktur_cell=edifact_struktur_cell,
                    middle_cell=middle_cell,
                    bedingung_cell=bedingung_cell,
                )

                ahb_table_row_dataframe = ahb_table_row.parse(row_type=current_row_type)

                if ahb_table_row_dataframe is not None:
                    ahb_table_dataframe = pd.concat([ahb_table_dataframe, ahb_table_row_dataframe], ignore_index=True)
            else:
                # this case covers the page break situation

                # check for conditions_text
                contains_condition_texts = any(paragraph.text != "" for paragraph in bedingung_cell.paragraphs)
                # conditions are always at the top of a dataelement
                # add condition texts
                if contains_condition_texts:
                    AhbSubTable.combine_condition_text(ahb_table_dataframe, bedingung_cell)

                # add new row regularly
                ahb_table_row = AhbTableRow(
                    seed=table_meta_data,
                    edifact_struktur_cell=edifact_struktur_cell,
                    middle_cell=middle_cell,
                    bedingung_cell=bedingung_cell,
                )
                ahb_table_row_dataframe = ahb_table_row.parse(row_type=current_row_type)

                # look at first line to determine if it is broken
                first_paragraph = middle_cell.paragraphs[0]

                if ahb_table_row_dataframe is not None:
                    if AhbSubTable.is_broken_line(
                        table=ahb_table_dataframe,
                        table_meta_data=table_meta_data,
                        paragraph=first_paragraph,
                    ):
                        AhbSubTable.add_broken_line(ahb_table_dataframe, ahb_table_row_dataframe)
                        # we have a broken line
                        ahb_table_dataframe = pd.concat(
                            [ahb_table_dataframe, ahb_table_row_dataframe.iloc[1:]],
                            ignore_index=True,
                        )
                    else:
                        ahb_table_dataframe = pd.concat(
                            [ahb_table_dataframe, ahb_table_row_dataframe],
                            ignore_index=True,
                        )

            # An AhbSubTable can span over two pages.
            # But after every page break, even if we're still in the same subtable,
            # there'll be the header at the top of every page.
            # This distracts our detection logic but we workaround it by remembering
            # the last two row types.
            table_meta_data.last_two_row_types[1] = table_meta_data.last_two_row_types[0]
            table_meta_data.last_two_row_types[0] = current_row_type
        return ahb_table_dataframe

    @staticmethod
    def initialize_dataframe_with_columns(columns: list[str]) -> pd.DataFrame:
        """
        Initialize a new dataframe with the given columns
        """
        return pd.DataFrame(
            columns=columns,
            dtype="str",
        )

    @classmethod
    def from_table_with_header(cls, docx_table: DocxTable) -> "AhbSubTable":
        """
        Create a new AhbSubTable instance from a docx table WITH header
        """

        ahb_table_meta_data = Seed.from_table(docx_table=docx_table)

        ahb_table_dataframe = AhbSubTable.initialize_dataframe_with_columns(columns=ahb_table_meta_data.column_headers)

        ahb_table_dataframe = cls._parse_docx_table(
            table_meta_data=ahb_table_meta_data,
            ahb_table_dataframe=ahb_table_dataframe,
            docx_table=docx_table,
        )

        return cls(table_meta_data=ahb_table_meta_data, table=ahb_table_dataframe)

    @classmethod
    def from_headless_table(cls, tmd: Seed, docx_table: DocxTable) -> "AhbSubTable":
        """
        Create a new AhbSubTable instance from a docx table WITHOUT header
        """

        ahb_table_dataframe = AhbSubTable.initialize_dataframe_with_columns(columns=tmd.column_headers)

        ahb_table_dataframe = cls._parse_docx_table(
            table_meta_data=tmd,
            ahb_table_dataframe=ahb_table_dataframe,
            docx_table=docx_table,
        )

        return cls(table_meta_data=tmd, table=ahb_table_dataframe)

    @staticmethod
    def iter_visible_cells(row: _Row) -> Generator[_Cell, None, None]:
        """
        This function makes sure that you will iterate over the cells you see in the word document.
        For more information go to https://github.com/python-openxml/python-docx/issues/970#issuecomment-877386927
        """
        table_row = row._tr  # pylint:disable=protected-access
        for table_column in table_row.tc_lst:
            yield _Cell(table_column, row.table)

    @staticmethod
    def add_text_to_last_row(ahb_table_dataframe: pd.DataFrame, row_index: int, column_index: int, text: str) -> None:
        """Add a text to the last row of the dataframe."""
        starts_with_known_suffix = any(text.startswith(suffix + " ") for suffix in KNOW_SUFFIXES)
        if len(text) > 0:
            if len(ahb_table_dataframe.iat[row_index, column_index]) > 0 and not starts_with_known_suffix:
                text = " " + text
            ahb_table_dataframe.iat[row_index, column_index] += text

    @staticmethod
    def add_broken_line(ahb_table_dataframe: pd.DataFrame, broken_line: pd.DataFrame) -> None:
        """Add a broken line to the dataframe."""
        for col_index in range(INDEX_OF_CODES_AND_QUALIFIER_COLUMN, len(ahb_table_dataframe.columns)):
            AhbSubTable.add_text_to_last_row(
                ahb_table_dataframe, ahb_table_dataframe.index.max(), col_index, str(broken_line.iat[0, col_index])
            )

    @staticmethod
    def combine_condition_text(ahb_table_dataframe: pd.DataFrame, bedingung_cell: _Cell) -> None:
        """Add the condition text to the dataframe."""
        conditions_text = " " + " ".join(
            paragraph.text for paragraph in bedingung_cell.paragraphs if paragraph.text != ""
        )
        last_valid_row = ahb_table_dataframe["Bedingung"].last_valid_index()
        conditions_text = ahb_table_dataframe.at[last_valid_row, "Bedingung"] + conditions_text
        # remove existing text
        ahb_table_dataframe.at[last_valid_row, "Bedingung"] = ""
        # remove remaining text to avoid misplacements
        for paragraph in bedingung_cell.paragraphs:
            paragraph.text = ""
        bedingung_cell.paragraphs[-1].text = conditions_text

    @staticmethod
    def is_broken_line(
        table: pd.DataFrame,
        table_meta_data: Seed,
        paragraph: Paragraph,
    ) -> bool:
        """
        Check for broken lines in the middle cell.
        """
        tabsplit_text = paragraph.text.split("\t")

        loc: Union[int, slice, NDArray[np.bool_]] = table.columns.get_loc("Beschreibung")

        # Ensure loc is an int
        if isinstance(loc, int):
            beschreibung_index: int = loc
        else:
            raise ValueError("The location of the column 'Beschreibung' is not an integer.")

        is_empty_middle_line = all(text == "" for text in tabsplit_text)
        is_broken_code_qualifier = (
            paragraph.paragraph_format.left_indent is not None
            and paragraph.paragraph_format.left_indent != table_meta_data.middle_cell_left_indent_position
            and table.iat[-1, beschreibung_index] != ""
            and table.iloc[-1, beschreibung_index + 1 :].ne("").any()
        )
        if is_broken_code_qualifier and len(tabsplit_text) == 1:
            # only broken code / qualifier
            assert (
                table.iat[-1, beschreibung_index] != "" and table.iloc[-1, beschreibung_index + 1 :].ne("").any()
            ), "no condition expected in broken line"
        there_are_conditions = (
            len(tabsplit_text) > 1
            and paragraph.paragraph_format.left_indent != table_meta_data.middle_cell_left_indent_position
        )

        return is_empty_middle_line or there_are_conditions or is_broken_code_qualifier



"""
A collection of functions to get information from AHB tables.
"""

from typing import Generator, Optional, Tuple, TypeGuard, Union

from docx.document import Document
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P
from docx.table import Table, _Cell
from docx.text.paragraph import Paragraph
from efoli import EdifactFormat, get_format_of_pruefidentifikator

from kohlrahbi.ahbtable.ahbcondtions import AhbConditions
from kohlrahbi.ahbtable.ahbpackagetable import AhbPackageTable
from kohlrahbi.ahbtable.ahbsubtable import AhbSubTable
from kohlrahbi.ahbtable.ahbtable import AhbTable
from kohlrahbi.logger import logger
from kohlrahbi.seed import Seed


def get_all_paragraphs_and_tables(parent: Union[Document, _Cell]) -> Generator[Union[Paragraph, Table], None, None]:
    """
    Yield each paragraph and table child within *parent*, in document order.
    Each returned value is an instance of either Table or Paragraph.
    *parent* would most commonly be a reference to a main Document object, but
    also works for a _Cell object, which itself can contain paragraphs and tables.
    """
    # pylint: disable=protected-access
    if isinstance(parent, Document):
        parent_elm = parent.element.body
    elif isinstance(parent, _Cell):
        parent_elm = parent._tc
    else:
        raise ValueError("Passed parent argument must be of type Document or _Cell")

    for child in parent_elm.iterchildren():
        if isinstance(child, CT_P):
            yield Paragraph(child, parent)
        elif isinstance(child, CT_Tbl):
            yield Table(child, parent)


def table_header_starts_with_text_edifact_struktur(table: Table) -> bool:
    """
    Check if the table header starts with the text "EDIFACT Struktur".
    """
    return table.cell(row_idx=0, col_idx=0).text.strip() == "EDIFACT Struktur"


def is_item_header_of_change_history_section(item: Paragraph | Table | None, style_name: str) -> TypeGuard[Paragraph]:
    """
    Checks if the given item is a header of the change history section.
    """
    # checking the style is quite expensive for the CPU because it includes some xpath searches;
    # we should only check the style if the other (easier/cheap) checks returned True so it at least
    return isinstance(item, Paragraph) and "Änderungshistorie" in item.text and "Heading" in style_name


def is_item_text_paragraph(item: Paragraph | Table | None, style_name: str) -> TypeGuard[Paragraph]:
    """
    Checks if the given item is a text paragraph.
    """
    return isinstance(item, Paragraph) and "Heading" not in style_name


def is_item_table_with_pruefidentifikatoren(item: Paragraph | Table | None) -> TypeGuard[Table]:
    """
    Check if the item is a Table and contains Pruefidentifikatoren.

    Args:
    item (Paragraph | Table | None): The item to check.

    Returns:
    bool: True if the item is a Table and contains Pruefidentifikatoren, False otherwise.
    """
    return isinstance(item, Table) and table_header_starts_with_text_edifact_struktur(table=item)


def is_item_headless_table(
    value: tuple[Union[Paragraph, Table, None], Union[AhbTable, None]],
) -> TypeGuard[tuple[Table, AhbTable]]:
    """
    Checks if the given item is a headless table.

    Args:
        item (Paragraph | Table | None): The item to be checked.
        seed (Seed): The seed object.
        ahb_table (AhbTable): The AhbTable object.

    Returns:
        bool: True if the item is a headless table, False otherwise.
    """
    item, ahb_table = value
    # return isinstance(item, Table) and seed is not None and ahb_table is not None
    return isinstance(item, Table) and ahb_table is not None


def get_ahb_table(document: Document, pruefi: str) -> Optional[AhbTable]:
    """
    Reads a docx file and extracts all information for a given Prüfidentifikator.
    If the Prüfidentifikator is not found or we reach the end of the AHB document
    - indicated by the section 'Änderungshistorie' - it returns None.

    Args:
        document: AHB word document which is read by python-docx package
        pruefi (str): The Prüfidentifikator to search for

    Returns:
        AhbTable or None: The extracted AHB table or None if not found
    """

    ahb_table: AhbTable | None = None
    seed: Seed | None = None
    searched_pruefi_is_found = False

    for item in get_all_paragraphs_and_tables(document):
        style_name = get_style_name(item)

        if is_item_text_paragraph(item, style_name):
            continue

        if reached_end_of_document(style_name, item):
            log_end_of_document(pruefi)
            break

        seed = update_seed(item, seed)

        if should_end_search(pruefi, seed, searched_pruefi_is_found):
            log_end_of_ahb_table(pruefi)
            break

        searched_pruefi_is_found, ahb_table = process_table(item, pruefi, searched_pruefi_is_found, ahb_table, seed)

    if ahb_table is not None:
        ahb_table.sanitize()
        return ahb_table

    log_pruefi_not_found(pruefi)
    return None


def get_style_name(item: Paragraph | Table) -> str:
    """Extracts and normalizes the style name of a document item."""
    return item.style.name if item.style else "None"  # type:ignore[no-any-return]


def reached_end_of_document(style_name: str, item: Paragraph | Table | None) -> bool:
    """Checks if the current item marks the end of the document."""
    return is_item_header_of_change_history_section(item, style_name)


def update_seed(item: Paragraph | Table | None, seed: Seed | None) -> Seed | None:
    """Updates the seed if the current item is a table with Prüfidentifikatoren."""
    if is_item_table_with_pruefidentifikatoren(item):
        return Seed.from_table(docx_table=item)
    return seed


def should_end_search(pruefi: str, seed: Seed | None, searched_pruefi_is_found: bool) -> bool:
    """Determines if the search for the AHB table should end."""
    if seed is None:
        return False
    return pruefi not in seed.pruefidentifikatoren and searched_pruefi_is_found


def process_table(
    item: Paragraph | Table | None,
    pruefi: str,
    searched_pruefi_is_found: bool,
    ahb_table: AhbTable | None,
    seed: Seed | None = None,
) -> tuple[bool, AhbTable]:
    """Processes tables to find and build the AHB table."""
    if is_item_table_with_pruefidentifikatoren(item):
        seed = Seed.from_table(docx_table=item)
        # pylint:disable=unsupported-membership-test
        if pruefi in seed.pruefidentifikatoren and not searched_pruefi_is_found:
            log_found_pruefi(pruefi)
            ahb_sub_table = AhbSubTable.from_table_with_header(docx_table=item)
            ahb_table = AhbTable.from_ahb_sub_table(ahb_sub_table=ahb_sub_table)
            searched_pruefi_is_found = True

    elif is_item_headless_table((item, ahb_table)):
        assert ahb_table is not None
        assert isinstance(item, Table)
        assert seed is not None
        ahb_sub_table = AhbSubTable.from_headless_table(docx_table=item, tmd=seed)
        ahb_table.append_ahb_sub_table(ahb_sub_table=ahb_sub_table)
    # actually, the ahb_table is none here (see test_kohlrahbi_cli_with_valid_arguments)
    return searched_pruefi_is_found, ahb_table  # type:ignore[return-value]


# Logging functions
def log_end_of_document(pruefi: str) -> None:
    """
    Logs that the end of the document was reached before finding the table for a given Prüfi.
    """
    logger.info("Reached the end of the document before finding the table for Prüfi '%s'.", pruefi)


def log_end_of_ahb_table(pruefi: str) -> None:
    """
    Logs that the end of the AHB table was reached for a given Prüfi.
    """
    logger.info("Reached the end of the AHB table for Prüfi '%s'.", pruefi)


def log_found_pruefi(pruefi: str) -> None:
    """
    Logs that the AHB table for a given Prüfi was found.
    """
    logger.info("Found the AHB table for Prüfi '%s'.", pruefi)


def log_pruefi_not_found(pruefi: str) -> None:
    """
    Logs that the Prüfi was not found in the provided document.
    """
    logger.warning("Prüfi '%s' was not found in the provided document.", pruefi)


# pylint: disable=too-many-branches
def get_all_conditions_from_doc(
    document: Document, edifact_format: EdifactFormat
) -> Tuple[Optional[AhbPackageTable], AhbConditions]:
    """
    Go through a given document and grasp all conditions and package tables for a given format.
    """
    package_table: AhbPackageTable
    conditions_table: AhbConditions
    package_tables: list[Table] = []
    conditions_tables: list[Table] = []
    seed = None

    # Iterate through the whole word document
    logger.info("🔁 Start iterating through paragraphs and tables")
    found_package_table = False
    for item in get_all_paragraphs_and_tables(document):
        style_name = get_style_name(item)
        if isinstance(item, Paragraph) and "Änderungshistorie" in item.text and "Heading" in style_name:
            logger.info(
                "Reached the end of the document, i.e. the section 'Änderungshistorie'.",
            )
        if is_item_text_paragraph(item, style_name):
            continue
        # processing of package tables
        if is_item_package_heading(item, style_name, edifact_format):
            found_package_table = True
            logger.info("🏁 Found Package Table for %s", edifact_format)
        elif isinstance(item, Table) and found_package_table:
            package_tables.append(item)
        elif found_package_table and not isinstance(item, Table):
            logger.info("We reached the end of the package table.")
            found_package_table = False

        if reached_end_of_document(style_name, item):
            log_end_of_document(edifact_format)

            break
        # processing of conditions tables
        seed = update_seed(item, seed)
        if is_relevant_pruefi_table(item, seed, edifact_format):
            conditions_tables.append(item)
        if is_last_row_unt_0062(item):
            seed = None
    if len(conditions_tables) > 0:
        conditions_table = AhbConditions.from_docx_table(conditions_tables, edifact_format)
    else:
        logger.warning("⛔️ No conditions found in the provided file.\n")

    if len(package_tables) > 0:
        package_table = AhbPackageTable.from_docx_table(package_tables)
        package_table.provide_packages(edifact_format)
    else:
        logger.warning("⛔️ No package table found in the provided file.\n")
    return package_table, conditions_table


def is_last_row_unt_0062(item: Table | Paragraph) -> bool:
    """
    Checks if the given table contains UNT segment in last row.
    """
    return isinstance(item, Table) and "UNT\t0062" == item.cell(row_idx=-1, col_idx=0).text.strip()


def is_relevant_pruefi_table(
    item: Paragraph | Table, seed: Seed | None, edifact_format: EdifactFormat
) -> TypeGuard[Table]:
    """compares new pruefis to last pruefi and thus checks whether new table"""
    return (
        isinstance(item, Table)
        and seed is not None
        and is_pruefi_of_edifact_format(seed.pruefidentifikatoren, edifact_format)
    )


def is_pruefi_of_edifact_format(last_pruefis: list[str], edifact_format: EdifactFormat) -> bool:
    """Checks if the pruefi is of the given edifact format"""
    return len(last_pruefis) > 0 and all(
        get_format_of_pruefidentifikator(pruefi) is edifact_format for pruefi in last_pruefis
    )


def is_item_package_heading(item: Paragraph | Table | None, style_name: str, edifact_format: EdifactFormat) -> bool:
    """
    Checks if the given item is the header of the package table.
    """
    return isinstance(item, Paragraph) and (
        (
            (style_name == "Heading 1")
            and "Übersicht der Pakete in der" in item.text
            and f"{edifact_format.name}" in item.text
        )
        or (((style_name == "Heading 2") and f"Übersicht der Pakete in der {edifact_format.name}" in item.text))
    )



"""
This module contains the UnfoldedAhbTable class.
"""

import copy
import json
import re
from functools import lru_cache
from pathlib import Path
from typing import Optional, Union
from uuid import uuid4

import pandas as pd
from efoli import EdifactFormat, get_format_of_pruefidentifikator
from more_itertools import first_true, peekable
from pydantic import BaseModel

from kohlrahbi.ahbtable.ahbtable import AhbTable, _column_letter_width_mapping
from kohlrahbi.logger import logger
from kohlrahbi.models.anwendungshandbuch import AhbLine, AhbMetaInformation, FlatAnwendungshandbuch
from kohlrahbi.models.flat_ahb_reader import FlatAhbCsvReader
from kohlrahbi.unfoldedahb.unfoldedahbline import UnfoldedAhbLine
from kohlrahbi.unfoldedahb.unfoldedahbtablemetadata import UnfoldedAhbTableMetaData

_segment_group_pattern = re.compile(r"^SG\d+$")
_segment_id_pattern = re.compile(r"^\d{5}$")


def _lines_are_equal_when_ignoring_guid(line1: AhbLine, line2: AhbLine) -> bool:
    """
    returns true iff the line1 and line2 are equal except for their guid
    """
    line1_copy = copy.deepcopy(line1)
    line2_copy = copy.deepcopy(line2)
    line1_copy.guid = None
    line2_copy.guid = None
    return line1_copy == line2_copy


@lru_cache
def _split_data_element_and_segment_id(value: str | None) -> tuple[str | None, str | None]:
    """
    returns the data element id and segment id
    """
    if not value:  # covers both None and empty string
        return None, None
    datenelement_id: str | None
    segment_id: str | None
    if _segment_id_pattern.match(value):
        datenelement_id = None
        segment_id = value
    else:
        datenelement_id = value
        segment_id = None
    return datenelement_id, segment_id


def _keep_guids_of_unchanged_lines_stable(
    updated_ahb: FlatAnwendungshandbuch, existing_ahb: FlatAnwendungshandbuch
) -> None:
    """
    Modifies the instance of updated_ahb such that the guids of all lines that are unchanged are the same as in the
    existing_ahb. Only applies if metadata of both AHBs match.
    """
    if updated_ahb.meta == existing_ahb.meta:
        existing_ahb_search_start_index = 0
        for update_index, updated_line in enumerate(updated_ahb.lines.copy()):
            if existing_line_match := first_true(  # ⚠ performance wise this goes like O(n^2)
                existing_ahb.lines[existing_ahb_search_start_index:],
                pred=lambda x: _lines_are_equal_when_ignoring_guid(
                    x, updated_line  # pylint:disable=cell-var-from-loop
                ),
            ):
                updated_ahb.lines[update_index].guid = existing_line_match.guid
                # if we found a line match, we can start the next search at the next line in the next loop iteration
                existing_ahb_search_start_index = existing_ahb.lines.index(existing_line_match) + 1


class UnfoldedAhb(BaseModel):
    """
    The UnfoldedAhb contains one Prüfidentifikator.
    Some columns in the AHB documents contain multiple information in one column e.g. Segmentname and Segmentgruppe.
    The unfolded classes add new columns/attributes to avoid the duplication of information in one column.
    """

    meta_data: UnfoldedAhbTableMetaData
    unfolded_ahb_lines: list[UnfoldedAhbLine]

    # pylint: disable=too-many-locals
    @classmethod
    def from_ahb_table(cls, ahb_table: AhbTable, pruefi: str) -> "UnfoldedAhb":
        """
        This function creates an UnfoldedAhb from an AhbTable.
        """
        unfolded_ahb_lines: list[UnfoldedAhbLine] = []
        current_section_name: str = ""
        current_segment_id: Optional[str] = None
        # we need to peek one iteration in front of us
        iterable_ahb_table = peekable(ahb_table.table.iterrows())

        for index, series in enumerate(iterable_ahb_table):
            row = series[1]

            current_section_name = UnfoldedAhb._get_section_name(
                segment_gruppe_or_section_name=row["Segment Gruppe"], last_section_name=current_section_name
            )

            if UnfoldedAhb._is_section_name(ahb_row=row):
                current_segment_id = None
                continue

            if UnfoldedAhb._is_segment_group(ahb_row=row):
                value_pool_entry, description = FlatAhbCsvReader.separate_value_pool_entry_and_name(
                    row["Codes und Qualifier"], row["Beschreibung"]
                )
                unfolded_ahb_lines.append(
                    UnfoldedAhbLine(
                        index=index,
                        segment_name=current_section_name,
                        segment_gruppe=row["Segment Gruppe"] or None,
                        segment=row["Segment"] or None,
                        datenelement=row["Datenelement"] or None,
                        code=value_pool_entry,
                        qualifier="",
                        beschreibung=description,
                        bedingung_ausdruck=row[pruefi] or None,
                        bedingung=row["Bedingung"],
                        segment_id=current_segment_id,
                    )
                )

            if UnfoldedAhb._is_segment_opening_line(ahb_row=row):
                current_segment_id = row["Segment ID"] or None
                unfolded_ahb_lines.append(
                    UnfoldedAhbLine(
                        index=index,
                        segment_name=current_section_name,
                        segment_gruppe=row["Segment Gruppe"] or None,
                        segment=row["Segment"] or None,
                        datenelement=None,
                        code=None,
                        qualifier="",
                        beschreibung=None,
                        bedingung_ausdruck=row[pruefi] or None,
                        bedingung=row["Bedingung"],
                        segment_id=current_segment_id,
                    )
                )
                continue

            if UnfoldedAhb._is_just_segment(ahb_row=row):
                value_pool_entry, description = FlatAhbCsvReader.separate_value_pool_entry_and_name(
                    row["Codes und Qualifier"], row["Beschreibung"]
                )
                unfolded_ahb_lines.append(
                    UnfoldedAhbLine(
                        index=index,
                        segment_name=current_section_name,
                        segment_gruppe=row["Segment Gruppe"] or None,
                        segment=row["Segment"] or None,
                        datenelement=_split_data_element_and_segment_id(row["Datenelement"])[0],
                        segment_id=current_segment_id,
                        code=value_pool_entry,
                        qualifier="",
                        beschreibung=description,
                        bedingung_ausdruck=row[pruefi] or None,
                        bedingung=row["Bedingung"],
                    )
                )
                continue

            if UnfoldedAhb._is_dataelement(ahb_row=row):
                value_pool_entry, description = FlatAhbCsvReader.separate_value_pool_entry_and_name(
                    row["Codes und Qualifier"], row["Beschreibung"]
                )

                if _segment_group_pattern.match(row["Segment Gruppe"]):
                    segment_group_key = row["Segment Gruppe"]
                else:
                    segment_group_key = None

                unfolded_ahb_lines.append(
                    UnfoldedAhbLine(
                        index=index,
                        segment_name=current_section_name,
                        segment_gruppe=segment_group_key,
                        segment=row["Segment"] or None,
                        datenelement=row["Datenelement"] or None,
                        code=value_pool_entry,
                        qualifier="",
                        beschreibung=description,
                        bedingung_ausdruck=row[pruefi] or None,
                        bedingung=row["Bedingung"],
                        segment_id=current_segment_id,
                    )
                )
                continue
            if any(unfolded_ahb_lines) and UnfoldedAhb._is_just_value_pool_entry(ahb_row=row):
                unfolded_ahb_lines.append(
                    UnfoldedAhbLine(
                        index=index,
                        segment_name=current_section_name,
                        segment_gruppe=unfolded_ahb_lines[-1].segment_gruppe,
                        segment=unfolded_ahb_lines[-1].segment,
                        datenelement=unfolded_ahb_lines[-1].datenelement,
                        code=row["Codes und Qualifier"],
                        qualifier="",
                        beschreibung=row["Beschreibung"],
                        bedingung_ausdruck=row[pruefi] or None,
                        bedingung=row["Bedingung"],
                    )
                )

        metadata = {"name": "" or None, "communication_direction": "" or None}
        if len(ahb_table.metadata) > 0:
            metadata["name"] = ahb_table.metadata[0].name or None
            metadata["communication_direction"] = ahb_table.metadata[0].communication_direction or None

        return cls(
            unfolded_ahb_lines=unfolded_ahb_lines,
            meta_data=UnfoldedAhbTableMetaData(
                pruefidentifikator=pruefi,
                beschreibung=metadata["name"],
                kommunikation_von=metadata["communication_direction"],
            ),
        )

    @staticmethod
    def _get_section_name(segment_gruppe_or_section_name: str, last_section_name: str) -> str:
        """
        This function reads the section name if there is one.
        If the given string 'segment_gruppe_or_section_name' does not contain a section name,
        the 'last_section_name' will be returned.
        """
        if not (segment_gruppe_or_section_name.startswith("SG") or segment_gruppe_or_section_name == ""):
            return segment_gruppe_or_section_name
        return last_section_name

    @staticmethod
    def _is_section_name(ahb_row: pd.Series) -> bool:  # type:ignore[type-arg]
        """
        Checks if the current AHB row is a section name.
        It uses the same logic as the function 'line_contains_only_segment_gruppe'
        So to avoid duplicate code, this function just calls the other function.
        """
        return AhbTable.line_contains_only_segment_gruppe(ahb_row)

    @staticmethod
    def _is_segment_group(ahb_row: pd.Series) -> bool:  # type:ignore[type-arg]
        """Checks if the current AHB row is a segment group."""

        if _segment_group_pattern.match(ahb_row["Segment Gruppe"]) and not ahb_row["Segment"]:
            return True
        return False

    @staticmethod
    def _is_segment_opening_line(ahb_row: pd.Series) -> bool:  # type:ignore[type-arg]
        """Checks if the current AHB row is a segment opening line.
        Example:

        SG3    CTA                                         Muss    Muss    Muss
        SG3    CTA    3139    IC    Informationskontakt    X       X       X

        The first line in the example is a segment opening line
        """

        if ahb_row["Segment"] and not ahb_row["Datenelement"]:
            return True
        return False

    @staticmethod
    def _is_just_segment(ahb_row: pd.Series) -> bool:  # type:ignore[type-arg]
        """
        Checks if the given AHB row is a segment
        """

        if (
            _segment_group_pattern.match(ahb_row["Segment Gruppe"])
            and ahb_row["Segment"]
            and not ahb_row["Datenelement"]
        ):
            return True
        if ahb_row["Datenelement"] is not None and _segment_id_pattern.match(ahb_row["Datenelement"]):
            return True
        return False

    @staticmethod
    def _is_dataelement(ahb_row: pd.Series) -> bool:  # type:ignore[type-arg]
        """
        Checks if the given AHB row is a dataelement
        """
        if ahb_row["Datenelement"]:
            return True
        return False

    @staticmethod
    def _is_just_value_pool_entry(ahb_row: pd.Series) -> bool:  # type:ignore[type-arg]
        """
        Checks if the given AHB row contains only a value pool entry (w/o Segment (group) and data element)
        """
        return (
            (not ahb_row["Segment Gruppe"])
            and (not ahb_row["Segment"])
            and (not ahb_row["Datenelement"])
            and ahb_row["Codes und Qualifier"]
        )

    def convert_to_flat_ahb(self) -> FlatAnwendungshandbuch:
        """
        Converts the unfolded AHB to a flat AHB.
        """
        meta = AhbMetaInformation(
            pruefidentifikator=self.meta_data.pruefidentifikator,
            description=self.meta_data.beschreibung,
            direction=self.meta_data.kommunikation_von,
        )
        lines: list[AhbLine] = []

        for unfolded_ahb_line in self.unfolded_ahb_lines:
            if _line_is_flatahb_line(unfolded_ahb_line):
                lines.append(
                    AhbLine(
                        guid=uuid4(),
                        segment_group_key=unfolded_ahb_line.segment_gruppe,
                        segment_code=unfolded_ahb_line.segment,
                        data_element=unfolded_ahb_line.datenelement,
                        segment_id=unfolded_ahb_line.segment_id,
                        value_pool_entry=unfolded_ahb_line.code,
                        name=unfolded_ahb_line.beschreibung or unfolded_ahb_line.qualifier,
                        ahb_expression=unfolded_ahb_line.bedingung_ausdruck,
                        conditions=unfolded_ahb_line.bedingung,
                        section_name=unfolded_ahb_line.segment_name,
                        index=unfolded_ahb_line.index,
                    )
                )
        try:
            return FlatAnwendungshandbuch(meta=meta, lines=lines)
        except ValueError:
            logger.error(
                "Could not convert the unfolded AHB to a flat AHB for Prüfidentifikator '%s'",
                self.meta_data.pruefidentifikator,
            )
            raise

    def get_flatahb_json_file_path(self, output_directory_path: Path) -> Path:
        """
        returns the filepath to where the flat ahb json will be dumped when using dump_flatahb_json()
        raises a value error when the pruefidentifikator is not a valid one
        """
        edifact_format = self._get_format()
        flatahb_output_directory_path = output_directory_path / str(edifact_format) / "flatahb"
        file_path = flatahb_output_directory_path / f"{self.meta_data.pruefidentifikator}.json"
        return file_path

    def dump_flatahb_json(self, output_directory_path: Path) -> None:
        """
        Converts the unfolded AHB to a flat AHB and writes it to a json file.
        The file will be stored in the directory:
            'output_directory_path/<edifact_format>/flatahb/<pruefidentifikator>.json'
        """
        try:
            file_path = self.get_flatahb_json_file_path(output_directory_path)
        except ValueError:
            return
        flatahb_directory = file_path.parent
        flatahb_directory.mkdir(parents=True, exist_ok=True)
        flat_ahb = self.convert_to_flat_ahb()
        if file_path.exists():
            with open(file_path, "r", encoding="utf-8") as file:
                file_content = file.read()
                existing_flat_ahb = FlatAnwendungshandbuch.model_validate_json(file_content)
            _keep_guids_of_unchanged_lines_stable(flat_ahb, existing_flat_ahb)
        json_dict = flat_ahb.model_dump(mode="json")
        with open(file_path, "w", encoding="utf-8") as file:
            json.dump(json_dict, file, ensure_ascii=False, indent=2, sort_keys=True)
        logger.info(
            "The flatahb file for %s is saved at %s",
            self.meta_data.pruefidentifikator,
            file_path.absolute(),
        )
        del flat_ahb
        del json_dict
        if "existing_flat_ahb" in locals():
            del existing_flat_ahb

    def convert_to_dataframe(self) -> pd.DataFrame:
        """
        Converts the unfolded AHB to a pandas dataframe.
        """
        unfolded_ahb_lines = [
            {
                "Segmentname": unfolded_ahb_line.segment_name,
                "Segmentgruppe": unfolded_ahb_line.segment_gruppe,
                "Segment": unfolded_ahb_line.segment,
                "Datenelement": unfolded_ahb_line.datenelement,
                "Segment ID": unfolded_ahb_line.segment_id,
                "Code": unfolded_ahb_line.code,
                "Qualifier": unfolded_ahb_line.qualifier,
                "Beschreibung": unfolded_ahb_line.beschreibung,
                "Bedingungsausdruck": unfolded_ahb_line.bedingung_ausdruck,
                "Bedingung": unfolded_ahb_line.bedingung,
            }
            for unfolded_ahb_line in self.unfolded_ahb_lines
            if _line_is_flatahb_line(unfolded_ahb_line)
        ]

        df = pd.DataFrame(unfolded_ahb_lines)
        df.fillna(value="", inplace=True)
        return df

    def get_csv_file_path(self, output_directory_path: Path) -> Path:
        """
        returns the filepath to where the CSV will be dumped when using dump_csv()
        raises a value error when the pruefidentifikator is not a valid one
        """
        edifact_format = self._get_format()
        csv_output_directory_path = output_directory_path / str(edifact_format) / "csv"
        file_path = csv_output_directory_path / f"{self.meta_data.pruefidentifikator}.csv"
        return file_path

    def dump_csv(self, path_to_output_directory: Path) -> None:
        """
        Dump a UnfoldedAHB table into a csv file.
        The file will be stored in the directory:
            'path_to_output_directory/<edifact_format>/csv/<pruefidentifikator>.csv'
        """
        df = self.convert_to_dataframe()
        try:
            csv_file_path = self.get_csv_file_path(path_to_output_directory)
        except ValueError:
            return
        csv_output_directory_path = csv_file_path.parent
        csv_output_directory_path.mkdir(parents=True, exist_ok=True)

        df.to_csv(csv_file_path, encoding="utf-8")
        logger.info("The csv file for %s is saved at %s", self.meta_data.pruefidentifikator, csv_file_path.absolute())
        del df

    def _get_format(self) -> EdifactFormat:
        """extract the edifact format from the metadata pruefidentifikator"""
        edifact_format = get_format_of_pruefidentifikator(self.meta_data.pruefidentifikator)
        if edifact_format is None:
            logger.warning("'%s' is not a pruefidentifikator", self.meta_data.pruefidentifikator)
            raise ValueError(f"'{self.meta_data.pruefidentifikator}' is not a pruefidentifikator")
        return edifact_format

    def get_xlsx_file_path(self, output_directory_path: Path) -> Path:
        """
        returns the filepath to where the xlsx will be dumped when using dump_xlsx()
        raises a value error when the pruefidentifikator is not a valid one
        """
        edifact_format = self._get_format()
        xlsx_output_directory_path: Path = output_directory_path / str(edifact_format) / "xlsx"
        file_path = xlsx_output_directory_path / f"{self.meta_data.pruefidentifikator}.xlsx"
        return file_path

    def dump_xlsx(self, path_to_output_directory: Path) -> None:
        """
        Dump a AHB table of a given pruefi into an excel file.
        The file will be stored in the directory:
            'path_to_output_directory/<edifact_format>/xlsx/<pruefidentifikator>.xlsx'
        """
        try:
            excel_file_path = self.get_xlsx_file_path(path_to_output_directory)
        except ValueError:
            return
        xlsx_output_directory_path = excel_file_path.parent
        xlsx_output_directory_path.mkdir(parents=True, exist_ok=True)

        df = self.convert_to_dataframe()
        try:
            # https://github.com/PyCQA/pylint/issues/3060
            # pylint: disable=abstract-class-instantiated
            with pd.ExcelWriter(excel_file_path, engine="xlsxwriter") as writer:
                df.to_excel(writer, sheet_name=f"{self.meta_data.pruefidentifikator}")
                # pylint: disable=no-member
                workbook = writer.book
                worksheet = writer.sheets[f"{self.meta_data.pruefidentifikator}"]
                wrap_format = workbook.add_format({"text_wrap": True})
                for column_letter, column_width in _column_letter_width_mapping.items():
                    excel_header = f"{column_letter}:{column_letter}"
                    worksheet.set_column(excel_header, column_width, wrap_format)
                logger.info("💾 Saved file(s) for Pruefidentifikator %s", self.meta_data.pruefidentifikator)
        except PermissionError:
            logger.error("The Excel file %s is open. Please close this file and try again.", excel_file_path)

        logger.info(
            "The xlsx file for %s is saved at %s",
            self.meta_data.pruefidentifikator,
            excel_file_path.absolute(),
        )


def _line_is_flatahb_line(unfolded_ahb_line: UnfoldedAhbLine) -> bool:
    """
    returns true if the unfolded_ahb_line is a valid line for a flat AHB
    """
    return (
        unfolded_ahb_line.segment is not None or unfolded_ahb_line.segment_gruppe is not None
    ) and unfolded_ahb_line.bedingung_ausdruck is not None


def _get_ahb(ahb_model_or_path: Union[FlatAnwendungshandbuch, UnfoldedAhb, Path]) -> FlatAnwendungshandbuch:
    """
    returns the AHB model
    """
    if isinstance(ahb_model_or_path, FlatAnwendungshandbuch):
        return ahb_model_or_path
    if isinstance(ahb_model_or_path, UnfoldedAhb):
        return ahb_model_or_path.convert_to_flat_ahb()
    if isinstance(ahb_model_or_path, Path):
        with open(ahb_model_or_path, "r", encoding="utf-8") as file:
            file_content = file.read()
            return FlatAnwendungshandbuch.model_validate_json(file_content)
    raise ValueError(
        f"argument must be either a FlatAnwendungshandbuch, UnfoldedAhb or a Path but was {type(ahb_model_or_path)}"
    )


def are_equal_except_for_guids(
    ahb_1: Union[FlatAnwendungshandbuch, UnfoldedAhb, Path],
    ahb_2: Union[FlatAnwendungshandbuch, UnfoldedAhb, Path],
) -> bool:
    """returns true iff both provided AHBs are equal except for/when ignoring their line guids"""
    ahb1 = _get_ahb(ahb_1)
    ahb2 = _get_ahb(ahb_2)
    if ahb1.meta != ahb2.meta:
        return False
    if len(ahb1.lines) != len(ahb2.lines):
        return False
    for line1, line2 in zip(ahb1.lines, ahb2.lines, strict=True):
        if not _lines_are_equal_when_ignoring_guid(line1, line2):
            return False
    return True



"""
This module contains the enumeration for the colors of different row types
"""

from enum import StrEnum


class RowTypeColor(StrEnum):
    """
    Colors for each RowType in rgb hex notation
    """

    HEADER = "de324c"  #: e.g. "EDIFACT Struktur"
    SEGMENTNAME = "f4895f"  #: e.g. "Nachrichten-Kopfsegment"
    SEGMENTGRUPPE = "f8e16f"  #: e.g. "SG2"
    SEGMENT = "95cf92"  #: e.g. "    UNH" or "SG2  NAD"
    DATENELEMENT = "369acc"  #: e.g. "    UNH 0062"
    EMPTY = "9656a2"  #: e.g. ""



"""
This module provides the QualityMapTable class
"""

from pathlib import Path

import pandas as pd
from docx.table import Table
from pydantic import BaseModel, ConfigDict

from kohlrahbi.ahbtable.ahbsubtable import AhbSubTable
from kohlrahbi.logger import logger


class QualityMapTable(BaseModel):
    """
    This class contains the quality map table.

    Die SG für die Datenübermittlung haben in den Segmenten DE1153 SG6 RFF, DE1229 SG8 SEQ und DE3035 SG12 NAD
    verschiedene Codes, die die Qualität der enthaltenen Daten festlegen.
    Die Qualitätsstufen sind:
        •   Bestellte Daten: Konfiguration, die der Absender beim Empfänger bestellt.
        •   Gültige Daten: Verbindliche Stammdaten, die vom Verantwortlichen bereitgestellt und
                           vom Berechtigten übernommen werden müssen.
        •   Informative Daten: Datenstand des Absenders zum Zeitpunkt der Nachricht, ohne Gültigkeitszeitraum,
                               nur zur Information im entsprechenden Use-Case.
        •   Erwartete Daten: Erwartung des Berechtigten, nur für Abrechnungsdaten, Stammdatenänderungen und
                             Datenclearing relevant.
        •   Im System vorhandene Daten: Datenstand des Berechtigten, ausschließlich für das Datenclearing.
    """

    table: pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_docx_quality_map_table(cls, docx_table: Table) -> "QualityMapTable":
        """
        Create a QualityMapTable object from a quality map table.
        """

        quality_map_rows: list[list[str]] = []

        for row in docx_table.rows:
            sanitized_cells = list(AhbSubTable.iter_visible_cells(row=row))

            is_header_row = sanitized_cells[0].text.strip() == "Qualität\n\nSegmentgruppe"
            if is_header_row:
                continue
            quality_map_rows.append([cell.text for cell in sanitized_cells])

        headers = [
            "Qualität \\ Segmentgruppe",
            "Bestellte Daten",
            "Gültige Daten",
            "Informative Daten",
            "Erwartete Daten",
            "Im System vorhandene Daten",
        ]

        df = pd.DataFrame(quality_map_rows, columns=headers)

        return cls(table=df)

    def save_to_csv(self, output_path: Path) -> None:
        """
        Save the quality map table to a csv file.
        """
        self.table.to_csv(output_path, index=False, encoding="utf-8")
        logger.info("💾 Saved quality map table to '%s'", output_path)

    def save_to_xlsx(self, output_path: Path) -> None:
        """
        Save the quality map table to an xlsx file.
        """
        self.table.to_excel(output_path, index=False)
        logger.info("💾 Saved quality map table to '%s'", output_path)



"""
This module contains all functions to define the type of a row of the tables in an AHB.
"""

from docx.oxml.ns import qn
from docx.oxml.parser import OxmlElement
from docx.shared import RGBColor
from docx.table import _Cell

from kohlrahbi.enums import RowType


def set_table_header_bg_color(cell: _Cell, hex_color: str) -> _Cell:
    """
    set background shading for Header Rows
    """
    table_cell = cell._tc  # pylint:disable=protected-access
    table_cell_properties = table_cell.get_or_add_tcPr()
    cl_shading = OxmlElement("w:shd")
    cl_shading.set(qn("w:fill"), hex_color)  # Hex of Dark Blue Shade {R:0x00, G:0x51, B:0x9E}
    table_cell_properties.append(cl_shading)
    return cell


def is_row_header(edifact_struktur_cell: _Cell) -> bool:
    """Checks if the current row is a header.

    Args:
        edifact_struktur_cell (_Cell): Indicator cell

    Returns:
        bool:
    """
    if edifact_struktur_cell.text == "EDIFACT Struktur":
        return True

    return False


def is_row_segmentname(edifact_struktur_cell: _Cell) -> bool:
    """Checks if the current row contains just a segment name.
       Example: "Nachrichten-Kopfsegment"

    Args:
        edifact_struktur_cell (_Cell): Indicator cell

    Returns:
        bool:
    """
    try:
        colour_is_grey: bool = edifact_struktur_cell.paragraphs[0].runs[0].font.color.rgb == RGBColor(128, 128, 128)
        return colour_is_grey
    except IndexError:
        return False


def is_row_segmentgruppe(edifact_struktur_cell: _Cell, left_indent_position: int) -> bool:
    """Checks if the current row is a segmentgruppe.
       Example: "SG2"

    Args:
        edifact_struktur_cell (_Cell): Indicator cell
        left_indent_position (int): Position of the left indent

    Returns:
        bool:
    """
    return (
        edifact_struktur_cell.paragraphs[0].paragraph_format.left_indent != left_indent_position
        and "\t" not in edifact_struktur_cell.text
        and not edifact_struktur_cell.text == ""
    )


def is_row_segment(edifact_struktur_cell: _Cell, left_indent_position: int) -> bool:
    """Checks if the current row is a segment.
       Example: "UNH", "SG2\tNAD"

    Args:
        edifact_struktur_cell (_Cell): Indicator cell
        left_indent_position (int): Position of the left indent

    Returns:
        bool:
    """
    # |   UNH    |
    if (
        edifact_struktur_cell.paragraphs[0].paragraph_format.left_indent == left_indent_position
        and "\t" not in edifact_struktur_cell.text
        and not edifact_struktur_cell.text == ""
    ):
        return True

    # | SG2\tNAD |
    if (
        not edifact_struktur_cell.paragraphs[0].paragraph_format.left_indent == left_indent_position
        and edifact_struktur_cell.text.count("\t") == 1
    ):
        return True

    return False


def is_row_datenelement(edifact_struktur_cell: _Cell, left_indent_position: int) -> bool:
    """Checks if the current row is a datenelement.
       Example: "UNH\t00062", "SG2\tNAD\t3035"

    Args:
        edifact_struktur_cell (_Cell): Indicator cell
        left_indent_position (int): Position of the left indent

    Returns:
        bool:
    """
    # |   UNH\t0062 |
    if (
        edifact_struktur_cell.paragraphs[0].paragraph_format.left_indent == left_indent_position
        and "\t" in edifact_struktur_cell.text
    ):
        return True

    # | SG2\tNAD\t3035 |
    if (
        not edifact_struktur_cell.paragraphs[0].paragraph_format.left_indent == left_indent_position
        and edifact_struktur_cell.text.count("\t") == 2
    ):
        return True

    return False


def is_row_empty(edifact_struktur_cell: _Cell) -> bool:
    """Checks if the current row is empty.
       Example: ""
    Args:
        edifact_struktur_cell (_Cell): Indicator cell

    Returns:
        bool:
    """
    return edifact_struktur_cell.text == ""


def get_row_type(edifact_struktur_cell: _Cell, left_indent_position: int) -> RowType:
    """Defines the type of the current row.

    Args:
        edifact_struktur_cell (_Cell): Indicator cell
        left_indent_position (int): Position of the left indent

    Raises:
        NotImplemented: Gets raised if the RowType got not to be defined

    Returns:
        RowType: Type of the current row
    """
    if is_row_header(edifact_struktur_cell=edifact_struktur_cell):
        return RowType.HEADER

    if is_row_segmentname(edifact_struktur_cell=edifact_struktur_cell):
        return RowType.SEGMENTNAME

    if is_row_segmentgruppe(edifact_struktur_cell=edifact_struktur_cell, left_indent_position=left_indent_position):
        return RowType.SEGMENTGRUPPE

    if is_row_segment(edifact_struktur_cell=edifact_struktur_cell, left_indent_position=left_indent_position):
        return RowType.SEGMENT

    if is_row_datenelement(edifact_struktur_cell=edifact_struktur_cell, left_indent_position=left_indent_position):
        return RowType.DATENELEMENT

    if is_row_empty(edifact_struktur_cell=edifact_struktur_cell):
        return RowType.EMPTY

    raise NotImplementedError(f"Could not define row type of cell with text: {edifact_struktur_cell.text}")



"""
This module contains the class BedingungCell
"""

import re

import pandas as pd
from docx.table import _Cell
from pydantic import BaseModel, ConfigDict


class BedingungCell(BaseModel):
    """
    BedingungCell contains all information and a method
    to extract the Bedingungen of an AHB Bedingung cell.
    """

    table_cell: _Cell

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def parse(self, ahb_row_dataframe: pd.DataFrame) -> pd.DataFrame:
        """
        Parses a cell in the Bedingung column and puts the information into the appropriate column of the dataframe.
        """
        bedingung = self.table_cell.text
        bedingung = self.beautify_bedingungen(bedingung)

        row_index = ahb_row_dataframe.index.max()
        ahb_row_dataframe.at[row_index, "Bedingung"] += bedingung
        return ahb_row_dataframe

    # pylint: disable=line-too-long
    @staticmethod
    def beautify_bedingungen(bedingung: str) -> str:
        """
        Beautifies the Bedingungen by removing the given line breaks and insert the line breaks at the correct places.

        Example input:
        [494] Das hier genannte Datum muss der Zeitpunkt sein, zu dem das Dokument erstellt wurde, oder ein Zeitpunkt, der davor liegt\n[931] Format: ZZZ = +00
        result:
        [494] Das hier genannte Datum muss der Zeitpunkt sein, zu dem das Dokument erstellt wurde, oder ein Zeitpunkt, der davor liegt
        [931] Format: ZZZ = +00
        """
        beautified_bedingung = bedingung.replace("\n", " ")

        matches = re.findall(r"\[\d+\]", beautified_bedingung)
        for match in matches[1:]:
            index = beautified_bedingung.find(match)
            beautified_bedingung = beautified_bedingung[:index].rstrip() + "\n" + beautified_bedingung[index:]

        return beautified_bedingung.lstrip()
